"""
Functions for SV breakend clustering.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""
from collections import OrderedDict, defaultdict
from pybedtools import BedTool
from statistics import median
import math


# SV clustering main function
def sv_cluster(
        beddata,
        parse,
        buf,
        maxovl,
        mincov,
        contigs,
        hsb_switch,
        seed
):
    """Returns a dictionary of clusters to SV breakends (mm) or list of clustered SV breakends in strings (hsb)

Keyword arguments:
- beddata: BED information for all reads in input
- parse: parse data from breakpoint_parser
- buf: coordinate clustering buffer in base pairs
- maxovl: maximum number of read overlap allowed for breakend-supporting or opposing- reads
- mincov: minimum number of breakend-supporting reads required for a breakend
- contigs: list of contig names from input reference genome
- hsb_switch: boolean to switch between pre and post blast re-evaluation
- seed: seed for SV index numbering

    """
    # Parse SV locations and cluster them
    # If not blast alignment
    if not hsb_switch:
        cluster_parse = rangecollect(parse,
                                     buf,
                                     contigs,
                                     hsb_switch,
                                     mincov
                                     )
        return cluster_parse
    else:  # If blast alignment
        clustid2coord, clustid2reads, clustid2class, clustid2ins_size, infodict, classdict, svsizedict = rangecollect(parse,
                                                                                                                      buf,
                                                                                                                      contigs,
                                                                                                                      hsb_switch,
                                                                                                                      mincov
                                                                                                                      )

        # Convert beddata to BED format
        bed3 = normalbed(beddata)
        bed3 = BedTool('\n'.join(bed3), from_string=True)
        bed3 = bed3.sort()

        # Convert SV breakends to BED format
        # bed4 = svbed2(readteam, mainclass)
        bed4 = svbed2(clustid2reads, clustid2class, clustid2coord)
        bed4 = BedTool('\n'.join(bed4), from_string=True)
        bed4 = bed4.sort()

        # Intersect all read regions to SV breakends
        intersect = bed3.intersect(bed4, wa=True, wb=True)

        # Create a duplicated readteam dict where reads have their unique identifier removed
        # newdictnouniq = remove_uniq(readteam)
        newdictnouniq = remove_uniq(clustid2reads)

        # Parse intersection BED to discover breakend-opposing reads for each breakend
        svnormalcov = intersection3(intersect, clustid2reads, newdictnouniq, clustid2class)

        # Output all information into a list of strings
        output, seed = arrange(svnormalcov, clustid2coord, clustid2reads, maxovl, mincov, infodict, clustid2class, svsizedict, seed, clustid2ins_size)
        return output, seed + 1


# Function to collect information of each breakpoint entry and cluster breakpoints
def rangecollect(x, buf, contigs, hsb_switch, mincov):

    classdict, svsizedict, infodict = OrderedDict(), OrderedDict(), OrderedDict()
    leftclust, rightclust = OrderedDict(), OrderedDict()

    for c in contigs:
        leftclust[c] = OrderedDict()
        rightclust[c] = OrderedDict()

    for i in x:
        rnameidx = i.split('\t')[8]  # Read unique id
        infodict[rnameidx] = '\t'.join(i.split('\t')[0:10])  # Read info storage from total_out file
        svtype = str(svtypecorrector(i.split('\t')[3].split(' ')[0]))  # Get sv type and slightly modify it
        classdict[rnameidx] = svtype  # Assign sv type to classdict
        svsize = alpha(i.split('\t')[3].split(' ')[1].split('~')[0])  # Get sv size
        svsizedict[rnameidx] = svsize  # Assign sv size to svsizedict
        if len(i.split('\t')[6].split('~')) == 2:  # Not Inter translocation
            if svtype in ['Nov_Ins', 'bp_Nov_Ins']:  # If single breakend
                chm1 = i.split('\t')[6].split('~')[1].split(':')[0]  # Get contig name
                le = int(i.split('\t')[6].split('~')[1].split(':')[1].split('-')[0])  # Take first coord of Nov_Ins or bp_Nov_Ins
                col4 = i.split('\t')[3]  # Get SV type label
                # Assign coord to left or right group depending on strand
                if col4.split(' ')[0][0] == 'S':  # Head novel sequence
                    if col4.split(' ')[1].split('~')[1] == '+':
                        rightclust[chm1][rnameidx + '-l'] = le
                    elif col4.split(' ')[1].split('~')[1] == '-':
                        leftclust[chm1][rnameidx + '-l'] = le
                elif col4.split(' ')[0][0] == 'E':  # Tail novel sequence
                    if col4.split(' ')[1].split('~')[1] == '+':
                        leftclust[chm1][rnameidx + '-l'] = le
                    elif col4.split(' ')[1].split('~')[1] == '-':
                        rightclust[chm1][rnameidx + '-l'] = le
                elif col4.split(' ')[0][0] == 'N':  # Nov insertion with both breakends of insertion event
                    # Assign single coord to both left and right groups
                    leftclust[chm1][rnameidx + '-l'] = le
                    rightclust[chm1][rnameidx + '-r'] = le
                else:
                    raise Exception('Error: single SV breakend %s has unknown naming' % rnameidx)
            elif len(i.split('\t')[6].split('~')[1].split(':')[1].split('-')) == 2:  # If double breakends
                chm1 = i.split('\t')[6].split('~')[1].split(':')[0]  # Get contig name
                le = int(i.split('\t')[6].split('~')[1].split(':')[1].split('-')[0])  # Get default left coord
                r = int(i.split('\t')[6].split('~')[1].split(':')[1].split('-')[1])  # Get default right coord
                if svtype == 'TDupl':  # If tandem dup, reverse left and right groups
                    leftclust[chm1][rnameidx + '-r'] = r
                    rightclust[chm1][rnameidx + '-l'] = le
                elif svtype in ['Inv', 'Inv2']:
                    col4s = i.split('\t')[3].split(' ')[1].split('~')[1]  # Get SV strand info
                    if col4s.split(',')[0] == '+' and col4s.split(',')[1] == '-':
                        leftclust[chm1][rnameidx + '-l'] = le
                        leftclust[chm1][rnameidx + '-r'] = r
                    elif col4s.split(',')[0] == '-' and col4s.split(',')[1] == '+':
                        rightclust[chm1][rnameidx + '-l'] = le
                        rightclust[chm1][rnameidx + '-r'] = r
                    elif col4s.split(',')[0] == '+' and col4s.split(',')[1] == '+':
                        leftclust[chm1][rnameidx + '-l'] = le
                        rightclust[chm1][rnameidx + '-r'] = r
                    elif col4s.split(',')[0] == '-' and col4s.split(',')[1] == '-':
                        rightclust[chm1][rnameidx + '-l'] = le
                        leftclust[chm1][rnameidx + '-r'] = r
                    else:
                        raise Exception('Error: Inversion % has unidentified strand orientation' % rnameidx)
                elif svtype in ['Intra-Ins', 'Intra-Ins2']:
                    col4s = i.split('\t')[3].split(' ')[1].split('~')[1]  # Get SV strand info
                    if col4s.split(',')[0] == '+':
                        leftclust[chm1][rnameidx + '-l'] = le
                    elif col4s.split(',')[0] == '-':
                        rightclust[chm1][rnameidx + '-l'] = le
                    if col4s.split(',')[1] == '+':
                        rightclust[chm1][rnameidx + '-r'] = r
                    elif col4s.split(',')[1] == '-':
                        leftclust[chm1][rnameidx + '-r'] = r
                else:
                    leftclust[chm1][rnameidx + '-l'] = le
                    rightclust[chm1][rnameidx + '-r'] = r
        elif len(i.split('\t')[6].split('~')) == 3:  # Inter translocation
            chm1 = i.split('\t')[6].split('~')[1].split(':')[0]  # Get contig name1
            chm2 = i.split('\t')[6].split('~')[2].split(':')[0]  # Get contig name2
            col4s = i.split('\t')[3].split(' ')[1].split('~')[1]  # Get SV strand info
            le = int(i.split('\t')[6].split('~')[1].split(':')[1])  # Get default left coord
            r = int(i.split('\t')[6].split('~')[2].split(':')[1])  # Get default right coord
            # Assign coord to left or right group depending on strand
            if col4s.split(',')[0] == '+':
                leftclust[chm1][rnameidx + '-l'] = le
            elif col4s.split(',')[0] == '-':
                rightclust[chm1][rnameidx + '-l'] = le
            if col4s.split(',')[1] == '+':
                rightclust[chm2][rnameidx + '-r'] = r
            elif col4s.split(',')[1] == '-':
                leftclust[chm2][rnameidx + '-r'] = r
        else:
            raise Exception('Error: %s is missed' % rnameidx)
    
    leftchrnamelist, leftchrcoorlist = OrderedDict(), OrderedDict()
    rightchrnamelist, rightchrcoorlist = OrderedDict(), OrderedDict()
    
    # Place the coords and read ids in a linear list for clustering with the same coord ascending order
    for chrm in contigs:
        leftclust[chrm] = OrderedDict(sorted(leftclust[chrm].items(), key=lambda y: y[1]))
        rightclust[chrm] = OrderedDict(sorted(rightclust[chrm].items(), key=lambda y: y[1]))
        leftchrnamelist[chrm] = []
        leftchrcoorlist[chrm] = []
        rightchrnamelist[chrm] = []
        rightchrcoorlist[chrm] = []
        for key in leftclust[chrm]:
            leftchrnamelist[chrm].append(key)  # [:-2]
            leftchrcoorlist[chrm].append(leftclust[chrm][key])
        for key in rightclust[chrm]:
            rightchrnamelist[chrm].append(key)  # [:-2]
            rightchrcoorlist[chrm].append(rightclust[chrm][key])
    
    # Carry out clustering
    if not hsb_switch:
        cluster_parse = cluster(leftchrnamelist, leftchrcoorlist, rightchrnamelist, rightchrcoorlist, buf, svsizedict,
                                classdict, hsb_switch, mincov)
        return cluster_parse
    else:
        clustid2coord, clustid2reads, clustid2class, clustid2ins_size = cluster(leftchrnamelist, leftchrcoorlist, rightchrnamelist, rightchrcoorlist, buf, svsizedict,
                                                                        classdict, hsb_switch, mincov)
        return clustid2coord, clustid2reads, clustid2class, clustid2ins_size, infodict, classdict, svsizedict


"""
- cluster_parse: Dictionary where keys are the cluster coordinates + svtype and values are the breakend names respective to each 
                cluster.
- readteam: Dictionary where keys are cluster coordinates (e.g. chr1:100-chr1-200 or chr1:100 for Ins SVs) and values are 
            lists of reads with the respective cluster coordinates, with the first read in the list as the lead read.
- infodict: Dictionary of default data for each SV read from total_out
- classdict: Dictionary of the SV class of each read (key=read_id)
- mainclass: Dictionary where keys are cluster coordinates and values are the SV class of the lead read
- svsizedict: Dictionary of SV sizes for each SV (key=read_id)
"""


# Cluster coordinates
def cluster(leftchrnamelist,
            leftchrcoorlist,
            rightchrnamelist,
            rightchrcoorlist,
            buf,
            svsizedict,
            classdict,
            hsb_switch,
            mincov):

    leftcluster2read, rightcluster2read, read2cluster = OrderedDict(), OrderedDict(), OrderedDict()

    # Cluster coordinates for each contig using buf as reference, clustering separately for left and right groups
    # Note: Coordinates within buf bp of each other will be clustered together and averaged to obtain a final coord. This is
    # done regardless of SV type or validity. Hence, final clustered coord may be slightly different from coord given by each
    # read.
    # Left group clustering
    for chrm in leftchrnamelist:
        n = len(leftchrnamelist[chrm])
        last = None
        group = []
        groupnames = []
        for i in range(n):
            if last is None or leftchrcoorlist[chrm][i] - last <= buf:
                group.append(leftchrcoorlist[chrm][i])
                groupnames.append(leftchrnamelist[chrm][i])
            else:
                # avgcoord = int(round(sum(group) / len(group), 0))
                medcoord = int(median(group))
                leftcluster2read[chrm + ':' + str(medcoord)] = sorted(set(y[:-2] for y in groupnames))
                for read in sorted(set(groupnames)):
                    if read[:-2] not in read2cluster:
                        read2cluster[read[:-2]] = []
                    read2cluster[read[:-2]].append(chrm + ':' + str(medcoord) + 'l')
                group = [leftchrcoorlist[chrm][i]]
                groupnames = [leftchrnamelist[chrm][i]]
            last = leftchrcoorlist[chrm][i]
        if group:
            # avgcoord = int(round(sum(group) / len(group), 0))
            medcoord = int(median(group))
            leftcluster2read[chrm + ':' + str(medcoord)] = sorted(set(y[:-2] for y in groupnames))
            for read in sorted(set(groupnames)):
                if read[:-2] not in read2cluster:
                    read2cluster[read[:-2]] = []
                read2cluster[read[:-2]].append(chrm + ':' + str(medcoord) + 'l')
    
    # Right group clustering
    for chrm in rightchrnamelist:
        n = len(rightchrnamelist[chrm])
        last = None
        group = []
        groupnames = []
        for i in range(n):
            if last is None or rightchrcoorlist[chrm][i] - last <= buf:
                group.append(rightchrcoorlist[chrm][i])
                groupnames.append(rightchrnamelist[chrm][i])
            else:
                # avgcoord = int(round(sum(group) / len(group), 0))
                medcoord = int(median(group))
                rightcluster2read[chrm + ':' + str(medcoord)] = sorted(set(y[:-2] for y in groupnames))
                for read in sorted(set(groupnames)):
                    if read[:-2] not in read2cluster:
                        read2cluster[read[:-2]] = []
                    read2cluster[read[:-2]].append(chrm + ':' + str(medcoord) + 'r')
                group = [rightchrcoorlist[chrm][i]]
                groupnames = [rightchrnamelist[chrm][i]]
            last = rightchrcoorlist[chrm][i]
        if group:
            # avgcoord = int(round(sum(group) / len(group), 0))
            medcoord = int(median(group))
            rightcluster2read[chrm + ':' + str(medcoord)] = sorted(set(y[:-2] for y in groupnames))
            for read in sorted(set(groupnames)):
                if read[:-2] not in read2cluster:
                    read2cluster[read[:-2]] = []
                read2cluster[read[:-2]].append(chrm + ':' + str(medcoord) + 'r')
    """
    - leftcluster2read or rightcluster2read: Dict, key=cluster, value=list of read_ids having the cluster
    - read2cluster: Dict, key=read_id, value=cluster
    """
    
    # Links each of the two coord in an SV together by read information (Only if SV has two coord, else coord will link to [])
    clusterlink = OrderedDict()
    coordlist2 = []
    for read in read2cluster:
        if read2cluster[read][0] not in clusterlink:
            clusterlink[read2cluster[read][0]] = []
        if len(read2cluster[read]) == 1:  # If read only has single coord e.g. S-Nov_Ins or E-Nov_Ins, pass.
            pass
        elif len(read2cluster[read]) == 2:  # Append second coord to list
            clusterlink[read2cluster[read][0]].append(read2cluster[read][1])
            coordlist2.append(read2cluster[read][1])
        else:
            raise Exception("Error: Read %s has %d clusters" % (read, len(read2cluster[read])))
            # Read has 0 or more than 2 clusters

    # Filter 1
    # Clusters which have been regarded as second clusters will be removed as first clusters in dict to avoid double entry.
    for clusters in coordlist2:
        try:  # If second cluster is also a first cluster in clusterlink
            if not clusterlink[clusters]:  # If second cluster as a first cluster does not have own second cluster (single coord)
                del clusterlink[clusters]  # delete the cluster
        except KeyError:
            pass

    # Order second cluster lists chronologically
    for clusters in clusterlink:
        clusterlink[clusters] = sorted(set(clusterlink[clusters]))
    
    pastbps = {}
    clustid = 0
    clustid2coord = OrderedDict()
    clustid2reads = OrderedDict()
    clustid2class = OrderedDict()
    clustid2ins_size = OrderedDict()
    for coord in clusterlink:  # For clusters that are either unpaired or paired (left)
        if not clusterlink[coord]:  # If cluster is unpaired, which means single breakend SVs
            if coord[-1] == 'l':
                readlist = leftcluster2read[coord[:-1]]
            elif coord[-1] == 'r':
                readlist = rightcluster2read[coord[:-1]]
            else:
                raise Exception('Error: Coord %s did not specify direction' % coord)
            lead, main, med_size = leadread_bp(readlist, svsizedict, classdict)  # Gather lead read
            clustid2reads[clustid] = [lead]
            clustid2reads[clustid].extend(sorted(set(readlist).difference([lead])))  # Append remaining reads
            clustid2coord[clustid] = coord
            clustid2class[clustid] = main
            clustid2ins_size[clustid] = med_size
            clustid += 1
        else:  # If cluster is paired (left)
            if coord[-1] == 'l':
                reads1 = leftcluster2read[coord[:-1]]
            elif coord[-1] == 'r':
                reads1 = rightcluster2read[coord[:-1]]
            else:
                raise Exception('Error: Coord %s did not specify direction' % coord)
            for coord2 in clusterlink[coord]:
                if coord2[-1] == 'l':
                    reads2 = leftcluster2read[coord2[:-1]]
                elif coord2[-1] == 'r':
                    reads2 = rightcluster2read[coord2[:-1]]
                else:
                    raise Exception('Error: Coord %s did not specify direction' % coord2)
                if coord[-1] in ['l', 'r']:
                    coord = coord[:-1]
                coord2 = coord2[:-1]
                readlists, remainder = separate_cluster_type(reads1, reads2, classdict)                
                if readlists:
                    for _readlist in readlists:
                        readlist = [x for x in _readlist if x not in pastbps]
                        if readlist:
                            for i in readlist:
                                pastbps[i] = 1
                            lead, main, med_size = leadread(sorted(set(readlist)), svsizedict, classdict, hsb_switch)
                            # Check for reads exclusive to each left/right cluster and add them into readlist if conditions satisfied
                            for read in remainder:
                                if classdict[read] == 'bp_Nov_Ins':
                                    if read not in pastbps:
                                        readlist.append(read)
                                        pastbps[read] = 1
                            if main == 'Nov_Ins':
                                avg_coord = coord.split(':')[0] + ':' + \
                                            str(int(round((int(coord.split(':')[1]) + int(coord2.split(':')[1])) / 2, 0)))
                                clustid2reads[clustid] = [lead]
                                clustid2reads[clustid].extend(sorted(set(readlist).difference([lead])))  # Append remaining reads
                                clustid2coord[clustid] = avg_coord
                                clustid2class[clustid] = main
                                clustid2ins_size[clustid] = med_size
                                clustid += 1
                            else:  # Else include both coord into cluster label
                                clustid2reads[clustid] = [lead]
                                clustid2reads[clustid].extend(sorted(set(readlist).difference([lead])))  # Append remaining reads
                                clustid2coord[clustid] = coord + '-' + coord2
                                clustid2class[clustid] = main
                                clustid += 1
    if not hsb_switch:
        #return cluster_parse
        pass
    else:
        return clustid2coord, clustid2reads, clustid2class, clustid2ins_size

# Separate clusters by per read svtype (this would deprecate mainclasssv)
def separate_cluster_type(reads1, reads2, classdict):
    intersect = sorted(set(reads1).intersection(reads2))
    cluster_class_dict = {}
    remainder = []
    for read in intersect:
        svtype = classdict[read][0:3]  # Taking first 3 element of svtype (E.g. Inv2 -> Inv)
        if svtype != 'bp_':
            if svtype not in cluster_class_dict:  
                cluster_class_dict[svtype] = [read]
            else:
                cluster_class_dict[svtype].append(read)
        else:
            remainder.append(read)
    _readlists = []
    for svtype in cluster_class_dict:
        if len(cluster_class_dict[svtype]) >=2:
            _readlists.append(cluster_class_dict[svtype])
    for read in sorted(set(reads1).difference(reads2)) + sorted(set(reads2).difference(reads1)):
        remainder.append(read)
    return _readlists, remainder

def leadread(reads, svsizedict, classdict, hsb_switch):
    mainsvclass = mainclasssv(reads, classdict, hsb_switch)
    sizedict = OrderedDict()
    leader = ''
    sv_sizes = []
    for read in reads:
            if classdict[read] == mainsvclass:
                # sizedict[read] = svsizedict[read]
                sv_sizes.append(svsizedict[read])
    # sizedictsort = [key for (key, value) in sorted(sizedict.items(), key=lambda y: y[1], reverse=True)]
    # for read in sizedictsort:
    #     if classdict[read] == mainsvclass:
    #         leader = read
    #         break
    # med_size = int(median(sv_sizes))
    med_size = sorted(sv_sizes)[math.ceil((len(sv_sizes)+1)/2) - 1]
    # Change leader read to read with the median size instead of largest size
    for read in reads:
        if svsizedict[read] == med_size and classdict[read] == mainsvclass:
            leader = read
            break
    if leader == '':
        raise Exception("Error: Main SV class not found")
    # if mainsvclass == 'Nov_Ins':
    #     med_size = int(median(sv_sizes))
    # else:
    #     med_size = 0
    return leader, mainsvclass, med_size


def leadread_bp(reads, svsizedict, classdict):
    mainsvclass = 'bp_Nov_Ins'
    sizedict = OrderedDict()
    leader = ''
    sv_sizes = []
    for read in reads:
        # sizedict[read] = svsizedict[read]
        sv_sizes.append(svsizedict[read])
    # sizedictsort = [key for (key, value) in sorted(sizedict.items(), key=lambda y: y[1], reverse=True)]
    # for read in sizedictsort:
    #     if classdict[read] == mainsvclass:  # Read with the largest SV size and correspond to mainsvclass will be the leader
    #         leader = read
    #         break
    # med_size = int(median(sv_sizes))
    med_size = sorted(sv_sizes)[math.ceil((len(sv_sizes)+1)/2) - 1]
    # Change leader read to read with the median size instead of largest size
    for read in reads:
        if svsizedict[read] == med_size and classdict[read] == mainsvclass:
            leader = read
            break
    if leader == '':
        raise Exception("Error: Main SV class not found for reads %s" % ','.join(reads))
    # med_size = int(median(sv_sizes))
    return leader, mainsvclass, med_size


# Function to rank and filter best sv type
def mainclasssv(reads, classdict, hsb_switch):
    if hsb_switch:
        tier3sv = ['Inv2', 'Inter-Ins', 'Intra-Ins2']
        tierxsv = ['Intra-Ins', 'Inter']
        tier2sv = ['Inv', 'Del', 'Nov_Ins', 'TDupl']
        tier1sv = ['bp_Nov_Ins']
    else:
        tier3sv = ['Inv2', 'Inter-Ins', 'Intra-Ins2']
        tierxsv = ['Intra-Ins', 'Inter']
        tier2sv = ['Inv', 'Del', 'TDupl', 'Nov_Ins']
        tier1sv = ['bp_Nov_Ins']
    tmpdict = defaultdict(int)
    for read in reads:
        svclass = classdict[read]
        tmpdict[svclass] += 1
    if any(y in tier3sv for y in tmpdict):
        for s in tierxsv:
            try:
                del tmpdict[s]
            except KeyError:
                pass
        for s in tier2sv:
            try:
                del tmpdict[s]
            except KeyError:
                pass
        try:
            del tmpdict["bp_Nov_Ins"]
        except KeyError:
            pass
    elif any(y in tierxsv for y in tmpdict):
        for s in tier2sv:
            try:
                del tmpdict[s]
            except KeyError:
                pass
        try:
            del tmpdict["bp_Nov_Ins"]
        except KeyError:
            pass
    elif any(y in tier2sv for y in tmpdict):
        try:
            del tmpdict["bp_Nov_Ins"]
        except KeyError:
            pass
    elif any(y in tier1sv for y in tmpdict):
        pass
    else:
        a = ','.join([y for y in tmpdict])
        raise Exception("Error: Unknown SV class %s, %s" % (a, ','.join(reads)))
    mainsvclass = [key for (key, value) in sorted(tmpdict.items(), key=lambda x: x[1], reverse=True)][0]
    return mainsvclass


# Function to generate BED from beddata
def normalbed(beddata):
    totalbed = []
    len_buf = 100  # Arbituary length deduction buffer for normal reads,
    for line in beddata:
        if int(line[1]) > 200:  # only allow len_buf if read start is 200 bp away from contig start
            coord1 = int(line[1]) + len_buf
        else:
            coord1 = int(line[1])
        # note that coord2 might be close to contig end but cannot be adjusted as in coord1 due to unknown contig length.
        coord2 = int(line[2]) - len_buf
        if coord2 - coord1 > 0:
            totalbed.append(line[0] + '\t' + str(coord1) + '\t' + str(coord2) + '\t' + line[3])
    return totalbed


# Create BED from SV breakends
def svbed2(clustid2reads, clustid2class, clustid2coord):
    totalsvbed = []
    for clustid in clustid2reads:
        svtype = clustid2class[clustid]
        c = [x.strip('lr') for x in clustid2coord[clustid].split('-')]
        totalsvbed.append(c[0].split(':')[0] + '\t' + c[0].split(':')[1] + '\t' + str(int(c[0].split(':')[1]) + 1) + '\t' +
                          str(clustid) + '\t' + svtype + '\t' + '1')
        if svtype in ['Nov_Ins', 'bp_Nov_Ins']:
            pass
        else:
            totalsvbed.append(c[1].split(':')[0] + '\t' + c[1].split(':')[1] + '\t' + str(int(c[1].split(':')[1]) + 1) + '\t' +
                              str(clustid) + '\t' + svtype + '\t' + '2')
    return totalsvbed


# Parse BED intersection output
def intersection3(bed, clustid2reads, newdictnouniq, clustid2class):
    norm_connect = defaultdict(set)
    norm_cov = defaultdict(int)
    for line in bed:
        clustid = line.fields[7]
        ind = line.fields[9]
        read = line.fields[3]
        if read not in newdictnouniq[int(clustid)]:
            norm_connect[clustid + '~' + ind].add(read)
    for clustid in clustid2reads:
        svtype = clustid2class[clustid]
        if svtype in ['Del', 'Inv', 'Inv2']:
            norm_cov[clustid] = int(round((len(norm_connect[str(clustid) + '~1']) + len(norm_connect[str(clustid) + '~2']))/2, 0))
        elif svtype == 'TDupl':
            norm_cov[clustid] = len(norm_connect[str(clustid) + '~1'].intersection(norm_connect[str(clustid) + '~2']))
        elif svtype in ['Nov_Ins', 'bp_Nov_Ins']:
            norm_cov[clustid] = len(norm_connect[str(clustid) + '~1'])
        elif svtype in ['Intra-Ins2', 'Intra-Ins', 'Inter', 'Inter-Ins']:
            norm_cov[clustid] = min(len(norm_connect[str(clustid) + '~1']), len(norm_connect[str(clustid) + '~2']))
        else:
            raise Exception('Error: SV type not recognised.')
    return norm_cov


# Remove unique identifier from read name
def remove_uniq(clustid2reads):
    newdictnouniq = {}
    for clustid in clustid2reads:
        newdictnouniq[clustid] = set()
        for read in clustid2reads[clustid]:
            newdictnouniq[clustid].add(read.rsplit('~', 1)[0])
    return newdictnouniq


# Parse info into output
def arrange(svnormalcov, clustid2coord, clustid2reads, maxovl, mincov, infodict, clustid2class, svsizedict, seed, clustid2ins_size):
    output = []
    n = seed
    for clustid in clustid2reads:
        svtype = clustid2class[clustid]
        lcov = countcov(clustid2reads[clustid])
        bestread = clustid2reads[clustid][0]
        # Filter 3
        if mincov <= lcov <= maxovl and svnormalcov[clustid] <= maxovl:
            if svtype in ['Inter', 'Inter-Ins']:
                _cluster = clusterorder(clustid2coord[clustid],
                                        infodict[bestread].split('\t')[6].split('~')[1].split(':')[0],
                                        infodict[bestread].split('\t')[6].split('~')[1].split(':')[1],
                                        'Inter')
                output.append(
                    '\t'.join(infodict[bestread].split('\t')[0:6]) + '\tnv_SV' + str(n) + '-' +
                    infodict[bestread].split('\t')[6].split('~')[0] +
                    '~' + _cluster.split('-')[0] + '~' + _cluster.split('-')[1] + '\t' +
                    '\t'.join(infodict[bestread].split('\t')[7:]) + '\t' + str(lcov) + '\t' +
                    ','.join(dotter(clustid2reads[clustid])) + '\t' + str(svnormalcov[clustid])
                )
            elif svtype in ['Intra-Ins2', 'Intra-Ins']:
                _cluster = clusterorder(clustid2coord[clustid],
                                        infodict[bestread].split('\t')[6].split('~')[1].split(':')[0],
                                        infodict[bestread].split('\t')[6].split('~')[1].split(':')[1].split('-')[0],
                                        'Intra')
                chrm_coord1 = _cluster.split('-')[0]
                coord2 = _cluster.split('-')[1].split(':')[1]
                output.append(
                    '\t'.join(infodict[bestread].split('\t')[0:6]) + '\tnv_SV' + str(n) + '-' +
                    infodict[bestread].split('\t')[6].split('~')[0] + '~' +
                    chrm_coord1 + '-' + coord2 + '\t' +
                    '\t'.join(infodict[bestread].split('\t')[7:]) + '\t' + str(lcov) + '\t' +
                    ','.join(dotter(clustid2reads[clustid])) + '\t' + str(svnormalcov[clustid])
                    )
            else:
                if len(clustid2coord[clustid].split('-')) == 1:
                    output.append(
                        # ins_size_mod(infodict[bestread].split('\t')[0:6], clustid2ins_size[clustid]) + '\tnv_SV' + str(n) + '-' +
                        # infodict[bestread].split('\t')[6].split('~')[0] + '~' + clustid2coord[clustid].strip('lr') + '-' +
                        # str(int(clustid2coord[clustid].strip('lr').split(':')[1]) + 1) + '\t' +
                        # '\t'.join(infodict[bestread].split('\t')[7:]) + '\t' + str(lcov) + '\t' +
                        # ','.join(dotter(clustid2reads[clustid])) + '\t' + str(svnormalcov[clustid])
                        '\t'.join(infodict[bestread].split('\t')[0:6]) + '\tnv_SV' + str(n) + '-' +
                        infodict[bestread].split('\t')[6].split('~')[0] + '~' + clustid2coord[clustid].strip('lr') + '-' +
                        str(int(clustid2coord[clustid].strip('lr').split(':')[1]) + 1) + '\t' +
                        '\t'.join(infodict[bestread].split('\t')[7:]) + '\t' + str(lcov) + '\t' +
                        ','.join(dotter(clustid2reads[clustid])) + '\t' + str(svnormalcov[clustid])
                    )
                elif len(clustid2coord[clustid].split('-')) == 2:
                    chrm = clustid2coord[clustid].split('-')[0].split(':')[0]
                    coord1 = str(min(int(clustid2coord[clustid].split('-')[0].split(':')[1]), int(clustid2coord[clustid].split('-')[1].split(':')[1])))
                    coord2 = str(max(int(clustid2coord[clustid].split('-')[0].split(':')[1]), int(clustid2coord[clustid].split('-')[1].split(':')[1])))
                    # if svtype == 'Nov_Ins':
                    #     output.append(
                    #         ins_size_mod(infodict[bestread].split('\t')[0:6], clustid2ins_size[clustid]) + '\tnv_SV' + str(n) + '-' +
                    #         infodict[bestread].split('\t')[6].split('~')[0] + '~' +
                    #         chrm + ':' + coord1 + '-' + coord2 + '\t' +
                    #         '\t'.join(infodict[bestread].split('\t')[7:]) + '\t' + str(lcov) + '\t' +
                    #         ','.join(dotter(clustid2reads[clustid])) + '\t' + str(svnormalcov[clustid])
                    #     )
                    # else:
                    output.append(
                        '\t'.join(infodict[bestread].split('\t')[0:6]) + '\tnv_SV' + str(n) + '-' +
                        infodict[bestread].split('\t')[6].split('~')[0] + '~' +
                        chrm + ':' + coord1 + '-' + coord2 + '\t' +
                        '\t'.join(infodict[bestread].split('\t')[7:]) + '\t' + str(lcov) + '\t' +
                        ','.join(dotter(clustid2reads[clustid])) + '\t' + str(svnormalcov[clustid])
                    )
                else:
                    raise Exception('Error: Cluster name error')
            n += 1
    return output, n


# Function to convert non-digits to zero
def alpha(n):
    if str(n).isdigit():
        return int(n)
    else:
        return int(0)


# Function to standardize names of SV types
def svtypecorrector(svtype):
    if svtype == 'S-Nov_Ins_bp':
        return 'bp_Nov_Ins'
    elif svtype == 'E-Nov_Ins_bp':
        return 'bp_Nov_Ins'
    elif svtype == 'Inter-Ins(1)':
        return 'Inter-Ins'
    elif svtype == 'Inter-Ins(2)':
        return 'Inter-Ins'
    elif svtype == 'InterTx':
        return 'Inter'
    elif svtype == 'Inv(1)':
        return 'Inv2'
    elif svtype == 'Inv(2)':
        return 'Inv2'
    elif svtype == 'Intra-Ins(1)':
        return 'Intra-Ins2'
    elif svtype == 'Intra-Ins(2)':
        return 'Intra-Ins2'
    else:
        return svtype


# Convert single element list to .
def dotter(readlist):
    if len(readlist) == 1:
        return '.'
    else:
        return readlist[1:]


# Function to count total number of unique reads in an overlaping breakpoint entry
def countcov(readlist):
    newlist = []
    for read in readlist:
        newlist.append(read.rsplit('~', 1)[0])
    return len(set(newlist))


# Order coord in cluster label according to default order
def clusterorder(_cluster, d_chrm1, d_coord1, mode):
    chrm1 = _cluster.split('-')[0].split(':')[0]
    coord1 = _cluster.split('-')[0].split(':')[1]
    chrm2 = _cluster.split('-')[1].split(':')[0]
    coord2 = _cluster.split('-')[1].split(':')[1]
    if mode == 'Intra':
        coords = [coord1, coord2]  # make cluster coords in to list
        index1 = (min([0, 1], key=lambda i: abs(int(coords[i]) - int(d_coord1))))
        if index1 == 0:
            index2 = 1
        else:
            index2 = 0
        new_cluster = chrm1 + ':' + str(coords[index1]) + '-' + chrm2 + ':' + str(coords[index2])
    elif mode == 'Inter':
        if chrm1 == chrm2:  # if same chrm
            coords = [coord1, coord2]
            index1 = (min(range(len(coords)), key=lambda i: abs(int(coords[i]) - int(d_coord1))))
            if index1 == 0:
                index2 = 1
            else:
                index2 = 0
            new_cluster = chrm1 + ':' + str(coords[index1]) + '-' + chrm2 + ':' + str(coords[index2])
        else:  # if chrm are different
            if chrm1 == d_chrm1:
                new_cluster = _cluster
            else:  # if chrm2 == d_chrm1
                new_cluster = _cluster.split('-')[1] + '-' + _cluster.split('-')[0]
    else:
        raise Exception('Error: wrong mode')
    return new_cluster


# Modify sv size in INS
# def ins_size_mod(line, size):
#     return '\t'.join(line[0:3] + [line[3].split(' ')[0] + ' ' + str(size) + '~' + line[3].split('~')[1]] + line[4:])
    

# Note: Some bps seen in parse_file would be missing
# in cluster_file because they are overlapped by another
# bp within the same read



"""
Functions for index and blast alignment.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
import logging
import distutils.spawn
from subprocess import Popen, PIPE, STDOUT


# Check paths and executables
def check_exe(path, exe):
    if path is None:
        if distutils.spawn.find_executable(exe):
            return exe
        else:
            logging.critical("Error: %s executable is not in PATH" % exe)
            raise Exception("Error: %s executable is not in PATH" % exe)
    else:
        if distutils.spawn.find_executable(path):
            return path
        else:
            logging.critical("Error: %s path do not exist" % path)
            raise Exception("Error: %s path do not exist" % path)


def log_subprocess(out):
    for line in iter(out.readline, ''):
        if line != '\n':
            logging.debug(line.strip())


# Create blast index
# def blast_index(ref, mdb):
#     process = Popen([mdb, '-in', ref, '-input_type', 'fasta', '-dbtype', 'nucl'], universal_newlines=True,
#                     stdout=PIPE, stderr=STDOUT)
#     with process.stdout:
#         log_subprocess(process.stdout)
#     # Process finished
#     exitcode = process.wait()
#     if exitcode != 0:
#         logging.critical("Error: makeblastdb failed")
#         raise Exception("Error: makeblastdb failed, see log")


# Create FMD index
# def fmd_index(ref, hsb):
#     process = Popen([hsb, 'index', ref], universal_newlines=True, stdout=PIPE, stderr=STDOUT)
#     with process.stdout:
#         log_subprocess(process.stdout)
#     exitcode = process.wait()
#     if exitcode != 0:
#         logging.critical("Error: hs-blastn index failed")
#         raise Exception("Error: hs-blastn index failed, see log")


# Create windowmasker counts
# def window_counts(ref, counts_path, wmk):
#     process = Popen([wmk, '-in', ref, '-infmt', 'blastdb', '-mk_counts', '-out', counts_path],
#                     universal_newlines=True,
#                     stdout=PIPE, stderr=STDOUT)
#     with process.stdout:
#         log_subprocess(process.stdout)
#     # Process finished
#     exitcode = process.wait()
#     if exitcode != 0:
#         logging.critical("Error: windowmasker counts failed")
#         raise Exception("Error: windowmasker counts failed, see log")


# Create counts binary
# def window_binary(counts_path, obinary_path, wmk):
#     process = Popen([wmk, '-in', counts_path, '-sformat', 'obinary', '-out', obinary_path, '-convert'],
#                     universal_newlines=True, stdout=PIPE, stderr=STDOUT)
#     with process.stdout:
#         log_subprocess(process.stdout)
#     exitcode = process.wait()
#     if exitcode != 0:
#         logging.critical("Error: windowmasker obinary failed")
#         raise Exception("Error: windowmasker obinary failed, see log")


# Check presence of indexes
# def check_index(ref, wk_dir, ref_name):
#     counts_path = os.path.join(wk_dir, ref_name + '.counts')
#     obinary_path = counts_path + '.obinary'
#     # Only check if index files exist, they might be empty
#     if os.path.isfile(ref + '.nsq') and os.path.isfile(ref + '.nhr') and os.path.isfile(ref + '.nin') and os.path.isfile(
#             counts_path) and os.path.isfile(obinary_path) and os.path.isfile(ref + '.bwt') and os.path.isfile(ref + '.header') \
#             and os.path.isfile(ref + '.sa') and os.path.isfile(ref + '.sequence'):
#         return True
#     else:
#         return False


# Make indexes
# def make_index(mode, ref, wk_dir, ref_name, mdb, wmk, hsb):
#     counts_path = os.path.join(wk_dir, ref_name + '.counts')
#     obinary_path = counts_path + '.obinary'
#     if mode:
#         logging.info('Building blast database')
#         blast_index(ref, mdb)
#         logging.info('Windowmasker counts')
#         window_counts(ref, counts_path, wmk)
#         logging.info('Windowmasker binary')
#         window_binary(counts_path, obinary_path, wmk)
#         logging.info('Make FMD-index')
#         fmd_index(ref, hsb)
#     else:
#         if os.path.isfile(ref + '.nsq') and os.path.isfile(ref + '.nhr') and os.path.isfile(ref + '.nin'):
#             # Only check if index files exist, they might be empty
#             logging.debug("Make blast index skipped")
#         else:
#             logging.info('Building blast database')
#             blast_index(ref, mdb)
#         if os.path.isfile(counts_path):  # Only check if files exist, they might be empty
#             logging.debug("Windowmasker counts skipped")
#         else:
#             logging.info('Windowmasker counts')
#             window_counts(ref, counts_path, wmk)
#         if os.path.isfile(obinary_path):  # Only check if files exist, they might be empty
#             logging.debug("Windowmasker obinary skipped")
#         else:
#             logging.info('Windowmasker binary')
#             window_binary(counts_path, obinary_path, wmk)
#         if os.path.isfile(ref + '.bwt') and os.path.isfile(ref + '.header') and os.path.isfile(ref + '.sa') and os.path.isfile(
#                 ref + '.sequence'):  # Only check if index files exist, they might be empty
#             logging.debug("Make hs-blastn FMD-index skipped")
#         else:
#             logging.info('Make FMD-index')
#             fmd_index(ref, hsb)


# Minimap2 alignment
def align_mm(ref, read, wk_dir, read_name, ref_name, threads, mm, data_type, st):
    out_path = os.path.join(wk_dir, '%s-%s-mm.bam' % (read_name, ref_name))
    if data_type == 'ont':
        x = 'x map-ont'
    elif data_type == 'pacbio-clr':
        x = 'x map-pb'
    elif data_type == 'pacbio-ccs':
        x = 'x map-hifi'
    else:
        x = ''
    process = Popen([mm + ' -t ' + str(threads) + ' -a' + x + ' ' + ref + ' ' + read + ' | ' + st +
                     ' view -Sb - -o ' + out_path],
                    universal_newlines=True, stdout=PIPE, stderr=STDOUT, shell=True, executable='/bin/bash')
    with process.stdout:
        log_subprocess(process.stdout)
    exitcode = process.wait()
    if exitcode != 0:
        logging.critical("Error: minimap2 alignment failed")
        raise Exception("Error: minimap2 alignment failed, see log")
    return ['minimap2 -t ' + str(threads) + ' -a' + x + ' ' + ref + ' ' + read + ' | samtools view -Sb - -o ' + out_path,
            out_path]


# HS-BLASTN alignment
# def align_hsb(ref, wk_dir, ref_name, threads, hsb, debug):
#     obinary_path = os.path.join(wk_dir, str(ref_name) + '.counts.obinary')
#     out_path = os.path.join(wk_dir, 'temp-%s-blast.tab' % ref_name)
#     read_path = os.path.join(wk_dir, 'temp2.fa')
#     process = Popen([hsb + ' align -db ' + ref + ' -window_masker_db ' + obinary_path + ' -query ' + read_path + ' -out '
#                      + out_path + ' -outfmt 6 -num_threads ' + str(threads) + ' -max_target_seqs 3 -gapopen 0 '
#                      '-gapextend 4 -penalty -3 -reward 2'],
#                     universal_newlines=True, stdout=PIPE, stderr=STDOUT, shell=True, executable='/bin/bash')
#     # cmd = process.args[0]
#     with process.stdout:
#         log_subprocess(process.stdout)
#     exitcode = process.wait()
#     if exitcode != 0:
#         logging.critical("Error: hs-blastn alignment failed")
#         raise Exception("Error: hs-blastn alignment failed, see log")
#     if not debug:  # Remove temp2.fa
#         os.remove(read_path)
#     return ["hs-blastn align -db " + ref + " -window_masker_db obinary_path -query " + read_path + " -out " + out_path +
#             " -outfmt 6 -num_threads " + str(threads) + " -max_target_seqs 3 -gapopen 0 -gapextend 4 -penalty -3 -reward 2",
#             out_path]



"""
Functions for NN model inferencing using Keras module.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
import numpy as np
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
from tensorflow.keras.models import load_model


# Neural network inferencing
def inference(cluster, parse, model):
    odict, normalcovratiodict, classdict, normalcov, nmapdict, nchrdict = getoverlap(cluster)
    newpdata = filterparse(parse, odict)
    nsigndict = scalefeature(newpdata, odict, normalcovratiodict, classdict)
    tmp = []
    tmpkey = []
    for key in nsigndict:
        tmp.append(nsigndict[key])
        tmpkey.append(key)
    readarray = np.array(tmp, dtype=np.float64)
    nnmodel = load_model(model, compile=False)
    predictions = nnmodel.predict(readarray, verbose=0)
    predlist = [float(x[0]) for x in predictions]
    probdict = {}
    for i in range(len(tmpkey)):
        probdict[tmpkey[i]] = predlist[i]
    nn_out = []
    for i in cluster:
        finalp = finalprob(i.split('\t')[8], i.split('\t')[11], probdict)
        corrected_p = str(round(np.tanh(0.4 * float(odict[i.split('\t')[8]])) * finalp * readratiofilter(
            normalcovratiodict[i.split('\t')[8]], normalcov[i.split('\t')[8]]), 4))
        nn_out.append(i + '\t' + corrected_p)  # * nmapchrfilter(nmapdict[i.split('\t')[8]])))
        # tanh(0.4B)*P where B = no. of breakend supporting reads and P = average confidence probability of breakend
        # This function alters output probability based on number of breakend supporting reads.
        # Visualize on fooplot.com
    nn_out_sort = sorted(nn_out, key=get_read_name)
    return nn_out_sort


# Additional read ratio filter post NN inference
def readratiofilter(ratio, ncov):
    if 0.65 <= ratio:
        if ncov < 6:
            score = np.exp(10 * (1 - ratio) - 3.5)
        else:
            score = np.exp(10 * (1 - ratio) - 4.5)
        return min(score, 1.0)
    else:
        return 1.0


# Additional multimapping and contig filter post NN inference
def nmapchrfilter(nmap):
    mapscore = max(-np.exp(0.3 * nmap - 10) + 1, 0)  # nmap cap at
    return mapscore


# Function for sorting list according to read name
def get_read_name(x):
    return x.split('\t')[8]


# For each breakpoint line, assign the long read coverage and normal cov ratio
def getoverlap(overlap_data):
    overlapdict = {}
    normalcovratio = {}
    classdict = {}
    normalcov = {}
    nchrdict = {}
    nmapdict = {}
    for i in overlap_data:
        overlapdict[i.split('\t')[8]] = float(i.split('\t')[10])
        normalcovratio[i.split('\t')[8]] = round(float(i.split('\t')[12])/(int(i.split('\t')[10]) + float(i.split('\t')[12])), 5)
        normalcov[i.split('\t')[8]] = float(i.split('\t')[12])
        classdict[i.split('\t')[8]] = [i.split('\t')[3].split(' ')[0], i.split('\t')[3].split(' ')[1].split('~')[0]]
        nmapdict[i.split('\t')[8]] = int(i.split('\t')[9].split(',')[10])
        nchrdict[i.split('\t')[8]] = int(i.split('\t')[9].split(',')[12])
        if i.split('\t')[11] != '.':
            for d in i.split('\t')[11].split(','):
                overlapdict[d] = float(i.split('\t')[10])
                normalcovratio[d] = float(i.split('\t')[12])/(int(i.split('\t')[10]) + float(i.split('\t')[12]))
                normalcov[d] = float(i.split('\t')[12])
                classdict[d] = [i.split('\t')[3].split(' ')[0], i.split('\t')[3].split(' ')[1].split('~')[0]]
    return overlapdict, normalcovratio, classdict, normalcov, nmapdict, nchrdict


# Filter parse data for >0 long read coverage
def filterparse(parse_data, overlapdict):
    newlist = []
    for i in parse_data:
        try:
            if float(overlapdict[i.split('\t')[8]]) > 0:
                newlist.append(i)
        except KeyError:
            pass
    return newlist


# SV class detector (Not in use)
# def svclass_new(sv, size):
#     if sv in ['S-Nov_Ins_bp', 'E-Nov_Ins_bp', 'Nov_Ins']:
#         if float(size) <= 50:
#             return [0.1, 0, 0, 0, 0, 0]
#         elif float(size) <= 100:
#             return [0.2, 0, 0, 0, 0, 0]
#         elif float(size) <= 200:
#             return [0.4, 0, 0, 0, 0, 0]
#         elif float(size) <= 400:
#             return [0.6, 0, 0, 0, 0, 0]
#         elif float(size) <= 1000:
#             return [0.8, 0, 0, 0, 0, 0]
#         else:
#             return [1, 0, 0, 0, 0, 0]
#     elif sv == 'Del':
#         if float(size) <= 50:
#             return [0, 0.1, 0, 0, 0, 0]
#         elif float(size) <= 100:
#             return [0, 0.2, 0, 0, 0, 0]
#         elif float(size) <= 200:
#             return [0, 0.4, 0, 0, 0, 0]
#         elif float(size) <= 400:
#             return [0, 0.6, 0, 0, 0, 0]
#         elif float(size) <= 1000:
#             return [0, 0.8, 0, 0, 0, 0]
#         else:
#             return [0, 1, 0, 0, 0, 0]
#     elif sv in ['Inv', 'Inv(1)', 'Inv(2)']:
#         return [0, 0, 1, 0, 0, 0]
#     elif sv == 'TDupl':
#         return [0, 0, 0, 1, 0, 0]
#     elif sv in ['Intra-Ins', 'Intra-Ins(1)', 'Intra-Ins(2)']:
#         return [0, 0, 0, 0, 1, 0]
#     elif sv in ['InterTx', 'Inter-Ins(1)', 'Inter-Ins(2)']:
#         return [0, 0, 0, 0, 0, 1]
#     else:
#         raise Exception('ERROR: SV class not recognised %s' % sv)


# SV class detector
def svclass(sv, size):
    if str(sv) == 'S-Nov_Ins_bp' or sv == 'E-Nov_Ins_bp' or sv == 'Nov_Ins' or sv == 'Del':
        if float(size) <= 50:
            return 0
        elif float(size) <= 100:
            return 0.2
        elif float(size) <= 200:
            return 0.4
        else:
            return 1
    else:
        return 1


# Collect features for each breakpoint line and rescale the features between 0 and 1
def scalefeature(parse_data, overlapdict, normalcovratio, classdict):
    signdict = {}
    for i in parse_data:
        signdict[i.split('\t')[8]] = []
        for s in i.split('\t')[9].split(','):
            signdict[i.split('\t')[8]].append(s.strip('(%)'))
        signdict[i.split('\t')[8]].append(overlapdict[i.split('\t')[8]])
        signdict[i.split('\t')[8]].append(normalcovratio[i.split('\t')[8]])
        signdict[i.split('\t')[8]].append(svclass(classdict[i.split('\t')[8]][0], classdict[i.split('\t')[8]][1]))
    items = list(signdict.items())
    normsigndict = {}
    new_tup = list(map(signnorm, items))
    for i in new_tup:
        normsigndict[i[0]] = i[1]
    return normsigndict


# Set upper limits and scale signature
def signnorm(tup):
    s = tup[1]
    s[0] = round(float(s[0]) / 100, 5)
    s[1] = round(float(s[1]) / 100, 5)
    s[2] = round(float(s[2]) / 100, 5)
    s[3] = round(float(s[3]) / 100, 5)
    s[4] = round(float(s[4]) / 100, 5)
    s[5] = min(float(s[5]), 1)
    s[6] = min(float(s[6]), 1)
    s[7] = round(min(float(s[7]), 2) / 2, 5)
    s[8] = round(min(float(s[8]), 2) / 2, 5)
    s[9] = round(min(float(s[9]), 4) / 4, 5)
    s[10] = round((min(float(s[10]), 18) - 1) / 18, 5)
    s[11] = round((min(float(s[11]), 6) - 1) / 6, 5)
    s[12] = round((min(float(s[12]), 4) - 1) / 4, 5)
    s[13] = round(min(float(s[13]), 8) / 8, 5)
    s[14] = round(float(s[14]) / 100, 5)
    s[15] = round(float(s[15]) / 100, 5)
    s[16] = round(min(float(s[16]), 0.5) / 0.5, 5)
    s[17] = round(min(float(s[17]), 0.5) / 0.5, 5)
    s[18] = round(min(float(s[18]), 0.5) / 0.5, 5)
    s[19] = round(min(float(s[19]), 0.5) / 0.5, 5)
    s[20] = round(min(float(s[20]), 10) / 10, 5)
    return tup[0], s


# Function to average out NN score
def finalprob(leadname, fellow, probdict):
    leadprob = probdict[leadname]
    prob = float(leadprob)
    if fellow != '.':
        fellows = fellow.split(',')
        n = len(fellows) + 1
        for i in fellows:
            prob += float(probdict[i])
    elif fellow == '.':
        n = 1
    else:
        raise Exception("Error: Breakend fellow symbol unrecognised")
    fprob = float(prob)/n
    return fprob


# Output SV read overlap partners
def svread_ovl(wk_dir, out_nn):
    data = open(os.path.join(wk_dir, 'sv_support_reads.tsv'), 'w')
    data.write('SV-ID\tSupporting_reads (readname~index1,readname~index2...)\n')
    for line in out_nn:
        svid = line.split('\t')[6].split('~')[0]
        data.write(svid + '\t' + line.split('\t')[8] + ',' + line.split('\t')[11] + '\n')
    data.close()



"""
Functions for parsing and verifying input parameters.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import sys
import argparse
from nanovar import __version__


# Parse input
def input_parser(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="NanoVar is a long-read structural variant (SV) caller.",
                                     formatter_class=argparse.RawTextHelpFormatter, usage=msg())  # RawDescriptionHelpFormatter)

    def restrict_float(f):
        f = float(f)
        if f < 0.05 or f > 0.50:
            raise argparse.ArgumentTypeError("%r not in range [0.05, 0.50]" % (f,))
        return f

    parser.add_argument("input", type=str,
                        metavar="[FASTQ/FASTA/BAM/CRAM]",
                        help="""Path to long reads or mapped BAM/CRAM file.
Formats: fasta/fa/fa.gzip/fa.gz/fastq/fq/fq.gzip/fq.gz/bam/cram""")

    parser.add_argument("ref", type=str,
                        metavar="[reference_genome]",
                        help="""Path to reference genome in FASTA. Genome indexes created 
will overwrite indexes created by other aligners such as bwa.""")

    parser.add_argument("dir", type=str,
                        metavar="[work_directory]",
                        help="""Path to work directory. Directory will be created 
if it does not exist.""")

    parser.add_argument("--cnv", type=str, metavar="hg38",
                        default=None,
                        help=argparse.SUPPRESS)
    # help="""Detects large genomic copy-number variations using CytoCAD (e.g. loss/gain of whole chromosomes). Only works with hg38 genome assembly. Please state 'hg38' [None]"""

    parser.add_argument("-x", "--data_type", type=str, metavar="str",
                        default='ont',
                        help="""Type of long-read data [ont]
ont - Oxford Nanopore Technologies
pacbio-clr - Pacific Biosciences CLR
pacbio-ccs - Pacific Biosciences CCS""")

    parser.add_argument("-f", "--filter_bed", type=str, metavar="file",
                        help="""BED file with genomic regions to be excluded [None]
(e.g. telomeres and centromeres) Either specify name of in-built 
reference genome filter (i.e. hg38, hg19, mm10) or provide full 
path to own BED file.""")

    parser.add_argument("--annotate_ins", type=str, metavar="str",
                        default=None,
                        help="""Enable annotation of INS with NanoINSight, 
please specify species of sample [None]
Currently supported species are:
'human', 'mouse', and 'rattus'.
""")

    parser.add_argument("-c", "--mincov", type=int, metavar="int",
                        default=4,
                        help="Minimum number of reads required to call a breakend [4]")

    parser.add_argument("-l", "--minlen", type=int, metavar="int",
                        default=25,
                        help="Minimum length of SV to be detected [25]")

    parser.add_argument("-p", "--splitpct", type=restrict_float, metavar="float",
                        default=0.05,
                        help="""Minimum percentage of unmapped bases within a long read 
to be considered as a split-read. 0.05<=p<=0.50 [0.05]""")

    parser.add_argument("-a", "--minalign", type=int, metavar="int",
                        default=200,
                        help="Minimum alignment length for single alignment reads [200]")

    parser.add_argument("-b", "--buffer", type=int, metavar="int",
                        default=50,
                        help="Nucleotide length buffer for SV breakend clustering [50]")

    parser.add_argument("-s", "--score", type=float, metavar="float",
                        default=1.0,
                        help="""Score threshold for defining PASS/FAIL SVs in VCF [1.0]
Default score 1.0 was estimated from simulated analysis. """)

    parser.add_argument("--homo", type=float, metavar="float",
                        default=0.75,
                        help="""Lower limit of a breakend read ratio to classify a homozygous state [0.75]
(i.e. Any breakend with homo<=ratio<=1.00 is classified as homozygous)""")

    parser.add_argument("--hetero", type=float, metavar="float",
                        default=0.35,
                        help="""Lower limit of a breakend read ratio to classify a heterozygous state [0.35]
(i.e. Any breakend with hetero<=ratio<homo is classified as heterozygous)""")

    parser.add_argument("--sv_bam_out", action='store_true',
                        help="""Outputs a BAM file containing only SV-supporting reads with 
their corresponding SV-ID(s) stored in the "nv" tag separated by comma.""")

    parser.add_argument("--debug", action='store_true',
                        help="Run in debug mode")

    # parser.add_argument("--force", action='store_true',
    #                     help="Run full pipeline (i.e. do not skip index generation)")

    parser.add_argument("-v", "--version", action='version',
                        version=__version__,
                        help="Show version and exit")

    parser.add_argument("-q", "--quiet", action='store_true',
                        help="Hide verbose")

    parser.add_argument("-t", "--threads", type=int, metavar="int",
                        default=1,
                        help="Number of available threads for use [1]")

    parser.add_argument("--model", type=str, metavar="path",
                        help="Specify path to custom-built model")

    parser.add_argument("--mm", type=str, metavar="path",
                        help="Specify path to 'minimap2' executable")

    parser.add_argument("--st", type=str, metavar="path",
                        help="Specify path to 'samtools' executable")

    parser.add_argument("--ma", type=str, metavar="path",
                        help="Specify path to 'mafft' executable for NanoINSight")

    parser.add_argument("--rm", type=str, metavar="path",
                        help="Specify path to 'RepeatMasker' executable for NanoINSight")

    # parser.add_argument("--mdb", type=str, metavar="path",
    #                     help="Specify path to 'makeblastdb' executable")

    # parser.add_argument("--wmk", type=str, metavar="path",
    #                     help="Specify path to 'windowmasker' executable")

    # parser.add_argument("--hsb", type=str, metavar="path",
    #                     help="Specify path to 'hs-blastn' executable")

    parser.add_argument("--pickle", action='store_true',
                        help=argparse.SUPPRESS)

    parser.add_argument("--archivefasta", action='store_true',
                        help=argparse.SUPPRESS)

    parser.add_argument("--blastout", type=str, metavar="path",
                        help=argparse.SUPPRESS)

    args = parser.parse_args(args)
    return args


# Check if fasta/fastq is gzip compressed
def gzip_check(path):
    twobytes = b'\x1f\x8b'
    with open(path, 'rb') as f:
        return f.read(2) == twobytes


# Custom usage message
def msg():
    return "nanovar [options] [FASTQ/FASTA/BAM/CRAM] [REFERENCE_GENOME] [WORK_DIRECTORY]"



"""
Extract alternative sequences

Copyright (C) 2024 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import logging
from pybedtools import BedTool
from Bio.Seq import Seq

def get_alt_seq(wk_dir, out_nn, ref_path):
    bed_str = ''
    ins_id_seq = {}
    ins_read_seq = get_ins_seq(wk_dir)
    for line in out_nn:
       sv_type = line.split('\t')[3].split(' ')[0]
       sv_id = line.split('\t')[6].split('~')[0]
       strand = line.split('\t')[3].split('~')[1].split(',')[0]
       bed_line = make_bed(line, sv_type, sv_id)
       bed_str += bed_line
       if sv_type in ['Nov_Ins', 'E-Nov_Ins_bp', 'S-Nov_Ins_bp']:
           ins_seq = ins_seq_except(ins_read_seq, line.split('\t')[8])
           if strand == '+':
               ins_id_seq[sv_id] = ins_seq
           elif strand == '-':
               ins_id_seq[sv_id] = str(Seq(ins_seq).reverse_complement())
           else:
               raise Exception('{} strand symbol invalid'.format(strand))
    bed = BedTool(bed_str, from_string=True)
    fasta = bed.sequence(fi=ref_path, nameOnly=True)
    alt_seq = {}
    with open(fasta.seqfn) as f:
        for r in f:
            alt_seq[r.strip('>\n')] = next(f).upper().strip()
    for i in ins_id_seq:
        alt_seq[i] = alt_seq[i] + ins_id_seq[i]
    return alt_seq

def make_bed(line, sv_type, sv_id):
    data = line.split('\t')
    chrm = data[6].split('~')[1].split(':')[0]
    if sv_type in ['Inter-Ins(1)', 'Inter-Ins(2)', 'InterTx']:
        # end = int(data[6].split('~')[1].split(':')[1])
        # start = end - 1  # Including Inter-Ins/InterTx might be causing issue #89
        return ''
    elif sv_type == 'Del':
        start = int(data[6].split('~')[1].split(':')[1].split('-')[0]) - 1
        end = int(data[6].split('~')[1].split(':')[1].split('-')[1])
    elif sv_type in ['Nov_Ins', 'E-Nov_Ins_bp', 'S-Nov_Ins_bp']:
        end = int(data[6].split('~')[1].split(':')[1].split('-')[0])
        start = end - 1
    else:  # ['Inv', 'Inv(1)', 'Inv(2)', 'TDupl', 'Intra-Ins', 'Intra-Ins(1)', 'Intra-Ins(2)']
        # end = int(data[6].split('~')[1].split(':')[1].split('-')[0])
        # start = end - 1 # Might be causing issue #89
        return ''
    start = max(start, 0)
    end = max(end, 1)
    bed_line = '\t'.join([chrm, str(start), str(end), sv_id]) + '\n'
    return bed_line

def get_ins_seq(wk_dir):
    with open(os.path.join(wk_dir, 'ins_seq.fa')) as f:
        ins_read_seq = {}
        for r in f:
            id = r.split('::')[0].strip('>')
            ins_read_seq[id] = next(f).strip()
        return ins_read_seq

def ins_seq_except(ins_read_seq, read_id):
    try:
        return ins_read_seq[read_id]
    except KeyError:  # Not found in ins_seq.fa, below score threshold
        return 'N'




"""
Generate SV-supporting reads BAM file. 

@author: asmaa, cytham

# Test
wk_dir = '.'
vcf_file = 'sample.nanovar.pass.vcf'
sv_supp_tsv_file = 'sv_support_reads.tsv'
in_bam_path = 'input.bam'
input_type = 'bam'
ref_path = 'ref.fa'
vcf = os.path.join(wk_dir, vcf_file)
sv_supp_tsv = os.path.join(wk_dir, sv_supp_tsv_file)
create_sv_supp_bam(vcf, sv_supp_tsv, in_bam_path, wk_dir, input_type, ref_path)

"""

import os
import pandas as pd
import pysam

def create_sv_supp_bam(vcf, sv_supp_tsv, in_bam_path, wk_dir, input_type, ref_path):
    pass_ids = parse_pass_sv(vcf)
    supp_dict = parse_supp_tsv(sv_supp_tsv, pass_ids)
    in_bam = input_type_func(in_bam_path, input_type, ref_path)
    out_bam = os.path.join(wk_dir, "sv_support_reads.bam")
    with pysam.AlignmentFile(out_bam, "wb", template=in_bam) as bam_out:
        for seg in in_bam:
            try:
                id = supp_dict[seg.query_name]  # If segment is SV-supporting read
                seg.set_tag('nv', ','.join(id), value_type='Z', replace=True)  # Set SV-ID tag
                bam_out.write(seg)
            except KeyError:
                pass
    bam_out.close()

def parse_supp_tsv(sv_sup_tsv, pass_ids):
    supp_df = pd.read_csv(sv_sup_tsv, sep='\t', header=0)
    supp_df.rename(columns = {supp_df.columns[1]: 'readID'}, inplace = True)
    supp_dict = {}
    for index, row in supp_df.iterrows():
        for _id in row['readID'].split(','):
            try:
                _ = pass_ids[row['SV-ID']]  # If passed SV-ID
                id = _id.split('~')[0]
                try:
                    supp_dict[id].add(row['SV-ID'])
                except KeyError:
                    supp_dict[id] = set()
                    supp_dict[id].add(row['SV-ID'])
            except KeyError:
                pass
    return supp_dict

def parse_pass_sv(vcf):
    pass_ids = {}
    with open(vcf) as v:
        for line in v:
            if not line.startswith('#'):
                pass_ids[line.split('\t')[2]] = 0
    return pass_ids

def input_type_func(file_path, input_type, ref_path):
    if input_type in ['bam', 'raw']:
        sam = pysam.AlignmentFile(file_path, "rb")
    elif input_type == 'cram':
        sam = pysam.AlignmentFile(file_path, "rc", reference_filename=ref_path)
    return sam



#!/usr/bin/env python3

"""
NanoVar

This is the main executable file of the program NanoVar.

Copyright (C) 2021 Tham Cheng Yong

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


__author__ = 'CY Tham'
import os
import sys
import time
import pysam
import shutil
import logging
import nanovar
import random
import pickle
import threading
from datetime import datetime
from nanovar import __version__, input_parser, gzip_check, fastx_valid, bed_valid, check_exe, align_mm
#from progress.spinner import Spinner
#import nanoinsight


def main():
    # Parse arguments
    args = input_parser()
    file_path = args.input
    ref_path = args.ref
    ref_name = os.path.basename(ref_path).rsplit('.', 1)[0]
    wk_dir = args.dir
    data_type = args.data_type
    genome_filter = args.filter_bed
    cnv = args.cnv
    annotate_ins = args.annotate_ins
    minlen = args.minlen
    splitpct = args.splitpct
    minalign = args.minalign
    mincov = args.mincov
    buff = args.buffer
    score_threshold = args.score
    homo_t = args.homo
    het_t = args.hetero
    threads = args.threads
    debug = args.debug
    quiet = args.quiet
    # force = args.force
    model_path = args.model
    filter_bed_dir = os.path.join(os.path.dirname(nanovar.__file__), 'gaps')
    ref_dir = os.path.join(os.path.dirname(nanovar.__file__), 'ref')
    mm = args.mm
    st = args.st
    ma = args.ma
    rm = args.rm
    # mdb = args.mdb
    # wmk = args.wmk
    # hsb = args.hsb
    pickle_bool = args.pickle
    archivefasta = args.archivefasta
    # blastout = args.blastout
    blastout = 'blastoff'
    nv_cmd = ' '.join(sys.argv)

    # Check model file
    if model_path is None:
        if data_type == 'ont':
            model_path = os.path.join(os.path.dirname(nanovar.__file__), 'model',
                                      'ANN.E100B400L3N12-5D0.4-0.2SGDsee11_het_gup_v1.h5')
        elif data_type == 'pacbio-clr':
            model_path = os.path.join(os.path.dirname(nanovar.__file__), 'model',
                                      'ANN.E100B400L3N12-5D0.4-0.2SGDsee11_het_clr_v1.h5')
        elif data_type == 'pacbio-ccs':
            model_path = os.path.join(os.path.dirname(nanovar.__file__), 'model',
                                      'ANN.E100B400L3N12-5D0.4-0.2SGDsee11_het_ccs_v1.h5')
        else:
            logging.critical("Error: Invalid data type given '%s'" % data_type)
            raise Exception("Error: Invalid data type given '%s'" % data_type)
    else:
        if data_type not in ['ont', 'pacbio-clr', 'pacbio-ccs']:
            logging.info("Invalid data type given '%s', but irrelevant due to custom-built model use." % data_type)

    # Check homo_t > het_t
    if homo_t <= het_t:
        raise Exception("Error: --homo threshold %s is less than or equal to --hetero threshold %s" % (str(homo_t), str(het_t)))

    # Check for required executables
    mm = check_exe(mm, 'minimap2')
    st = check_exe(st, 'samtools')
    # mdb = check_exe(mdb, 'makeblastdb')
    # wmk = check_exe(wmk, 'windowmasker')
    # hsb = check_exe(hsb, 'hs-blastn')

    # Pre-check if annotating INS with NanoINSight
    if annotate_ins:
        try:
            import nanoinsight
        except ImportError:
            raise Exception("Error: nanoinsight module not found. Please install by 'pip install nanoinsight'.")
        species = annotate_ins
        nanoinsight.check_args(species)
        mafft_exe = nanoinsight.check_exe(ma, 'mafft')
        repmask_exe = nanoinsight.check_exe(rm, 'RepeatMasker')

    # If additional CNV mode
    hg38_size_dict = {}
    if cnv:
        raise Exception("Error: CNV module presently not available on this version. Please remove '--cnv' option. Sorry, we are actively improving.")
        # # Check CytoCAD
        # try:
        #     import cytocad
        #     data_dir = os.path.join(os.path.dirname(cytocad.__file__), 'data')
        #     hg38_size_path = os.path.join(data_dir, 'hg38_sizes_main.bed')
        #     with open(hg38_size_path) as f:
        #         for bed in f:
        #             line = bed.split('\t')
        #             hg38_size_dict[line[0]] = int(line[2])
        # except ImportError:
        #     raise Exception("Error: cytocad module not found, please install CytoCAD - https://github.com/cytham/cytocad")
        # # Check tagore and rsvg
        # _ = check_exe(None, 'tagore')
        # _ = check_exe(None, 'rsvg-convert')
        # # Check hg38 build input
        # if cnv != 'hg38':
        #     raise Exception("Error: Reference genome build %s is not recognised. CytoCAD only supports build hg38." % cnv)

    # Observe verbosity
    if quiet:
        sys.stdout = open(os.devnull, 'w')

    # Assign threads
    threads_mm = threads

    # Print initiation message
    now = datetime.now()
    now_str = now.strftime("[%d/%m/%Y %H:%M:%S]")
    print(now_str, "- NanoVar-v%s started" % __version__)
    print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Checking integrity of input files - ', end='', flush=True)

    # Setup working directory
    if not os.path.exists(wk_dir):
        os.makedirs(wk_dir)
    if not os.path.exists(os.path.join(wk_dir, 'fig')):
        os.makedirs(os.path.join(wk_dir, 'fig'))

    # Setup up logging
    log_file = os.path.join(wk_dir, 'NanoVar-{:%d%m%y-%H%M}.log'.format(datetime.now()))
    logging.basicConfig(filename=log_file, level=logging.DEBUG, format='[%(asctime)s] - %(levelname)s - %(message)s',
                        datefmt='%d/%m/%Y %H:%M:%S')
    logging.info('Initialize NanoVar log file')
    logging.info('Version: NanoVar-%s' % __version__)
    logging.info('Command: %s' % nv_cmd)

    # Detect read or map file
    filename = os.path.basename(file_path)
    read_suffix = ['.fa', '.fq', '.fasta', '.fastq', '.fa.gzip', '.fq.gzip', '.fa.gz', '.fq.gz', '.fasta.gz', '.fastq.gz']
    bam_suffix = '.bam'
    cram_suffix = '.cram'
    contig_list = []
    if any(filename.lower().endswith(s) for s in read_suffix):
        input_name = os.path.basename(file_path).rsplit('.f', 1)[0]
        input_type = 'raw'
        # Test gzip compression and validates read file
        if gzip_check(file_path):
            # read_para = "<(zcat " + file_path + ")"
            fastx_check = fastx_valid(file_path, "gz")
        else:
            # read_para = "<(cat " + file_path + ")"
            fastx_check = fastx_valid(file_path, "txt")
        if fastx_check[0] == "Fail":
            logging.critical("Error: Input FASTQ/FASTA file is corrupted around line %s +/- 4" % str(fastx_check[1]))
            raise Exception("Error: Input FASTQ/FASTA file is corrupted around line %s +/- 4" % str(fastx_check[1]))
        elif fastx_check[0] == "Pass":
            logging.debug("Input FASTQ/FASTA file passed")
    elif filename.lower().endswith(bam_suffix):
        save = pysam.set_verbosity(0)  # Suppress BAM index missing warning
        sam = pysam.AlignmentFile(file_path, "rb")
        pysam.set_verbosity(save)  # Revert verbosity level
        try:
            assert sam.is_bam, "Error: Input BAM file is not a BAM file."
            input_name = os.path.basename(file_path).rsplit('.bam', 1)[0]
            input_type = 'bam'
            fastx_check = []
            # Get BAM contigs from header
            header = sam.header.to_dict()
            for h in header['SQ']:
                contig_list.append(h['SN'])
        except AssertionError:
            logging.critical("Error: Input BAM file is not a BAM file.")
            raise Exception("Error: Input BAM file is not a BAM file.")
    elif filename.lower().endswith(cram_suffix):
        save = pysam.set_verbosity(0)  # Suppress index missing warning
        sam = pysam.AlignmentFile(file_path, "rc", reference_filename=ref_path)
        pysam.set_verbosity(save)  # Revert verbosity level
        try:
            assert sam.is_cram, "Error: Input CRAM file is not a CRAM file."
            input_name = os.path.basename(file_path).rsplit('.cram', 1)[0]
            input_type = 'cram'
            fastx_check = []
            # Get CRAM contigs from header
            header = sam.header.to_dict()
            for h in header['SQ']:
                contig_list.append(h['SN'])
        except AssertionError:
            logging.critical("Error: Input CRAM file is not a CRAM file.")
            raise Exception("Error: Input CRAM file is not a CRAM file.")
    else:
        logging.critical("Error: Input file is not recognised, please ensure file suffix has '.fa' or '.fq' or '.bam' or '.cram'")
        raise Exception("Error: Input file is not recognised, please ensure file suffix has '.fa' or '.fq' or '.bam' or '.cram'")

    if gzip_check(ref_path):
        logging.critical("Error: Input reference file is gzipped, please unzip it")
        raise Exception("Error: Input reference file is gzipped, please unzip it")

    # Logging config info
    logging.info('Input file: %s' % file_path)
    logging.info('Read type: %s' % data_type)
    logging.info('Reference genome: %s' % ref_path)
    logging.info('Working directory: %s' % wk_dir)
    logging.info('minimap2 path: %s' % shutil.which(mm))
    logging.info('samtools path: %s' % shutil.which(st))
    logging.info('Model: %s' % model_path)
    logging.info('Filter file: %s' % genome_filter)
    logging.info('Minimum number of reads for calling a breakend: %s' % str(mincov))
    logging.info('Minimum SV len: %s' % str(minlen))
    logging.info('Mapping percent for split-read: %s' % str(splitpct))
    logging.info('Length buffer for clustering: %s' % str(buff))
    logging.info('Score threshold: %s' % str(score_threshold))
    logging.info('Homozygous read ratio threshold: %s' % str(homo_t))
    logging.info('Heterozygous read ratio threshold: %s' % str(het_t))
    logging.info('Number of threads: %s\n' % str(threads))
    if input_type == 'raw':
        logging.info('Total number of reads in FASTQ/FASTA: %s\n' % str(fastx_check[1]))
    elif input_type == 'bam' or input_type == 'cram':
        logging.info('Total number of reads in FASTQ/FASTA: -\n')
    logging.info('NanoVar started')

    from Bio import SeqIO
    from collections import OrderedDict
    # Process reference genome
    contig_len_dict = OrderedDict()
    total_gsize = 0
    for seq_record in SeqIO.parse(ref_path, "fasta"):
        contig_len_dict[seq_record.id] = len(seq_record)
        total_gsize += len(seq_record)

    # Check if contig sizes from BAM/CRAM is same as hg38 if CNV mode
    if cnv:
        for chrom in hg38_size_dict:
            if chrom in contig_len_dict:
                if contig_len_dict[chrom] != hg38_size_dict[chrom]:
                    raise Exception("Error: Contig %s in BAM/CRAM (%i) has different length in hg38 genome assembly (%i)" %
                                    (chrom, contig_len_dict[chrom], hg38_size_dict[chrom]))
            else:
                raise Exception("Error: hg38 contig %s is absent in BAM/CRAM" % chrom)

    # Check BAM/CRAM contigs in reference genome
    if input_type == 'bam' or input_type == 'cram':
        for c in contig_list:
            if c not in contig_len_dict:
                logging.critical("Error: Contig %s in BAM/CRAM is absent in reference genome" % c)
                raise Exception("Error: Contig %s in BAM/CRAM is absent in reference genome" % c)

    # Check contig id for invalid symbols
    contig_omit = checkcontignames(contig_len_dict)

    # Validate filter BED file
    if genome_filter is not None:
        if genome_filter == 'hg38':
            filter_path = os.path.join(filter_bed_dir, genome_filter + '_curated_filter_main.bed')
        elif genome_filter in ('hg19', 'mm10'):
            filter_path = os.path.join(filter_bed_dir, genome_filter + '_filter-B400.bed')
        else:
            filter_path = genome_filter
        if os.path.isfile(filter_path):
            if bed_valid(filter_path, contig_len_dict):
                logging.debug("Genome filter BED passed")
        else:
            logging.critical("Error: Genome filter BED %s is not found" % filter_path)
            raise Exception("Error: Genome filter BED %s is not found" % filter_path)
    else:
        filter_path = genome_filter

    print('Pass')

    # Update progress
    if len(contig_omit) > 0:
        print('Warning: The following contig id(s) contains invalid symbols [ ~ : - ] %s. Reads mapping to these contig(s) '
              'will be ignored.' % ', '.join(['>' + x for x in contig_omit]))
        logging.warning('Warning: The following contig id(s) contains invalid symbols [ ~ : - ] %s. Reads mapping to these '
                        'contig(s) will be ignored.' % ', '.join(['>' + x for x in contig_omit]))

    # Aligning using minimap2
    if input_type == 'raw':
        logging.info('Read alignment using minimap2')
        print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Read mapping using minimap2 - ', end='', flush=True)
        mma = align_mm(ref_path, file_path, wk_dir, input_name, ref_name, threads_mm, mm, data_type, st)
        bam_path = mma[1]
        save = pysam.set_verbosity(0)  # Suppress index missing warning
        sam = pysam.AlignmentFile(bam_path, "rb")
        pysam.set_verbosity(save)  # Revert verbosity level
        print('Done')
    elif input_type == 'bam' or input_type == 'cram':
        logging.info('Input BAM/CRAM file, skipping minimap2 alignment')
        mma = ['-', '']
        bam_path = file_path

    # Update progress
    print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Analyzing read alignments and detecting SVs - ', end='', flush=True)
    
    # Parse and detect SVs
    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    from nanovar.nv_characterize import VariantDetect
    logging.info('Parsing BAM and detecting SVs')
    run = VariantDetect(wk_dir, bam_path, splitpct, minalign, filter_path, minlen, buff, model_path,
                        total_gsize, contig_len_dict, score_threshold, file_path, input_name, ref_path, ref_name, mma[0],
                        mincov, homo_t, het_t, debug, contig_omit, cnv, nv_cmd, sam)
    run.bam_parse_detect()
    run.coverage_stats()
    print('Done')

    logging.info('Total number of mapped reads: %s\n' % str(run.maps))
    
    # Update progress
    print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Clustering SV breakends - ', end='', flush=True)

    # SV breakend clustering and extracting INS and INV SVs
    run.cluster_nn2()
    print('Done')

    # Analyze TE
    print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Correcting DUP and detecting TE - ', end='', flush=True)
    pysam.faidx(os.path.join(wk_dir, 'temp1.fa'))
    run.dup_te_detect(ref_dir, threads, mm, st, data_type) #

    # Write parse2 and cluster intermediate files
    run.write2file(add_out=[])
    print('Done')

    # Pickle objects
    if pickle_bool:
        #saveobjects = (wk_dir, bam_path, splitpct, minalign, filter_path, minlen, buff, model_path,
        #               total_gsize, contig_len_dict, score_threshold, file_path, input_name, ref_path, ref_name, mma[0],
        #               mincov, homo_t, het_t, debug, contig_omit, cnv, run.rlendict, run.total_subdata, run.total_out,
        #               run.out_nn, sub_run.total_out)
        run.sam = None # Remove pysam instance 
        with open(os.path.join(wk_dir, "nv_run.pkl"), "wb") as f:
            pickle.dump(run, f)
    
    # Update progress
    print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Generating VCF files and report - ', end='', flush=True)
    
    # Generating VCF and HTML report
    run.vcf_report()
    print('Done')

    # Annotate insertions
    if annotate_ins:
        print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Annotating INS with NanoINSight - ', end='', flush=True)
        logging.info('Annotating INS with NanoINSight')
        vcf = os.path.join(wk_dir, '%s.nanovar.pass.vcf' % input_name)
        sv_sup = os.path.join(wk_dir, 'sv_support_reads.tsv')
        ins_seq = os.path.join(wk_dir, 'ins_seq.fa')
        id_seq, fasta_dir = nanoinsight.create_fa(vcf, wk_dir, sv_sup, ins_seq)
        con_fasta, threads_per_job = nanoinsight.create_cons(vcf, wk_dir, fasta_dir, id_seq, threads, mafft_exe, batch_size=100, num_parallel_workers=5)
        nanoinsight.rep_annote(wk_dir, con_fasta, threads_per_job, species, repmask_exe)
        print('Done')

    # Output SV-supporting BAM
    if args.sv_bam_out:
        print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), '- Creating SV-supporting BAM - ', end='', flush=True)
        logging.info('Creating SV-supporting BAM')
        from .nv_supp_bam import create_sv_supp_bam
        vcf = os.path.join(wk_dir, '%s.nanovar.pass.vcf' % input_name)
        sv_sup = os.path.join(wk_dir, 'sv_support_reads.tsv')
        create_sv_supp_bam(vcf, sv_sup, bam_path, wk_dir, input_type, ref_path)
        print('Done')
    
    # Delete temporary fasta file
    if not archivefasta:
        f = os.path.join(wk_dir, 'temp1.fa')
        os.remove(f)
        f = os.path.join(wk_dir, 'temp1.fa.fai')
        os.remove(f)
    
    if not debug:
        f = os.path.join(wk_dir, 'temp2.fa')
        os.remove(f)
        f = os.path.join(wk_dir, '%s.nanovar.total.vcf' % input_name)
        os.remove(f)
        f = os.path.join(wk_dir, 'genome.sizes')
        os.remove(f)
    
    logging.info('NanoVar ended')
    now = datetime.now()
    now_str = now.strftime("[%d/%m/%Y %H:%M:%S]")
    print(now_str, "- NanoVar ended")

# Check contig name
def checkcontignames(contig_len_dict):
    contig_omit = {}
    for contig in contig_len_dict:
        if "~" in contig or ":" in contig or "-" in contig:
            contig_omit[contig] = [0, int(contig_len_dict[contig])]
    return contig_omit

if __name__ == "__main__":
    main()



"""
Functions for verifying files.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import gzip
import logging


def fastx_valid(reads, ext):
    c = 0
    lineskip = 0
    totalc = 0
    corrupted = 0
    # rlen_dict = {}
    if ext == "gz":
        with gzip.open(reads, 'rt') as f:
            line1n2 = [next(f) for x in range(2)]
            try:
                line3n4 = [next(f) for x in range(2)]
            except StopIteration:
                pass
            if "line3n4" in locals():
                if line1n2[0][0] == ">" and line3n4[0][0] == ">":
                    xformat = "fasta"
                    # rlen_dict[line1n2[0].split()[0][1:].strip()] = len(line1n2[1].strip())
                    # rlen_dict[line3n4[0].split()[0][1:].strip()] = len(line3n4[1].strip())
                    for line in f:
                        c += 1
                        if lineskip == 0:
                            if line[0] != ">":
                                corrupted = 1
                                break
                            rname = line.split()[0][1:].strip()
                            lineskip = 1
                        else:
                            # rlen_dict[rname] = len(line.strip())
                            lineskip = 0
                            if c == 100000:
                                totalc = totalc + c
                                c = 0
                elif line1n2[0][0] == "@" and line3n4[0][0] == "+":
                    xformat = "fastq"
                    # rlen_dict[line1n2[0].split()[0][1:].strip()] = len(line1n2[1].strip())
                    for line in f:
                        c += 1
                        if lineskip == 0:
                            if line[0] != "@":
                                corrupted = 1
                                break
                            rname = line.split()[0][1:].strip()
                            lineskip = 1
                        elif lineskip == 1:
                            # rlen_dict[rname] = len(line.strip())
                            lineskip = 2
                        elif lineskip == 2:
                            if line[0] != "+":
                                corrupted = 1
                                break
                            lineskip = 3
                        else:
                            lineskip += 1
                            if lineskip == 4:
                                lineskip = 0
                            if c == 100000:
                                totalc = totalc + c
                                c = 0
                else:
                    c = 1
                    corrupted = 1
            else:
                if line1n2[0][0] == ">":
                    xformat = "fasta"
                    # rlen_dict[line1n2[0].split()[0][1:].strip()] = len(line1n2[1].strip())
                    c = 0
                else:
                    c = 1
                    corrupted = 1
    elif ext == "txt":
        with open(reads) as f:
            line1n2 = [next(f) for x in range(2)]
            try:
                line3n4 = [next(f) for x in range(2)]
            except StopIteration:
                pass
            if "line3n4" in locals():
                if line1n2[0][0] == ">" and line3n4[0][0] == ">":
                    xformat = "fasta"
                    # rlen_dict[line1n2[0].split()[0][1:].strip()] = len(line1n2[1].strip())
                    # rlen_dict[line3n4[0].split()[0][1:].strip()] = len(line3n4[1].strip())
                    for line in f:
                        c += 1
                        if lineskip == 0:
                            if line[0] != ">":
                                corrupted = 1
                                break
                            rname = line.split()[0][1:].strip()
                            lineskip = 1
                        else:
                            # rlen_dict[rname] = len(line.strip())
                            lineskip = 0
                            if c == 100000:
                                totalc = totalc + c
                                c = 0
                elif line1n2[0][0] == "@" and line3n4[0][0] == "+":
                    xformat = "fastq"
                    # rlen_dict[line1n2[0].split()[0][1:].strip()] = len(line1n2[1].strip())
                    for line in f:
                        c += 1
                        if lineskip == 0:
                            if line[0] != "@":
                                corrupted = 1
                                break
                            rname = line.split()[0][1:].strip()
                            lineskip = 1
                        elif lineskip == 1:
                            # rlen_dict[rname] = len(line.strip())
                            lineskip = 2
                        elif lineskip == 2:
                            if line[0] != "+":
                                corrupted = 1
                                break
                            lineskip = 3
                        else:
                            lineskip += 1
                            if lineskip == 4:
                                lineskip = 0
                            if c == 100000:
                                totalc = totalc + c
                                c = 0
                else:
                    c = 1
                    corrupted = 1
            else:
                if line1n2[0][0] == ">":
                    xformat = "fasta"
                    # rlen_dict[line1n2[0].split()[0][1:].strip()] = len(line1n2[1].strip())
                    c = 0
                else:
                    c = 1
                    corrupted = 1
    # Summarize and count reads
    if corrupted == 0:
        if xformat == 'fasta':
            if c == 0:
                return ["Pass", int(((totalc + c) / 2) + 1)]  # [Status, read count]
            elif c % 2 == 0:
                return ["Pass", int(((totalc + c) / 2) + 2)]  # [Status, read count]
            else:
                return ["Fail", int(totalc + c) + 2]  # [Status, line error]
        elif xformat == 'fastq':
            if c % 4 == 0:
                return ["Pass", int(((totalc + c) / 4) + 1)]  # [Status, read count]
            else:
                return ["Fail", int(totalc + c) + 4]  # [Status, line error]
    elif corrupted == 1:
        return ["Fail", int(totalc + c) + 4]  # [Status, line error]


def bed_valid(filter_path, contig_len_dict):

    # Make contig BED dict
    contig_bed_dict = {}
    contig_list = []
    for cid in contig_len_dict:
        contig_bed_dict[cid] = [0, contig_len_dict[cid]]
        contig_list.append(cid)

    # Validate BED file to genome
    c = 0
    with open(filter_path) as bed:
        for line in bed:
            c += 1
            cid = line.split('\t')[0]
            start = line.split('\t')[1]
            end = line.split('\t')[2]
            if cid in contig_list:
                try:
                    start = int(start)
                    end = int(end)
                except TypeError:
                    logging.critical("Error: Genome filter BED non-integer at line %s" % str(c))
                    raise TypeError("Error: Genome filter BED non-integer at line %s" % str(c))
                if contig_bed_dict[cid][0] <= start and contig_bed_dict[cid][1] >= end:
                    if start <= end:
                        return True
                    else:
                        logging.critical("Error: Genome filter BED start>end at line %s" % str(c))
                        raise TypeError("Error: Genome filter BED start>end at line %s" % str(c))
                else:
                    logging.critical("Error: Genome filter BED outside contig boundaries at line %s" % str(c))
                    raise TypeError("Error: Genome filter BED outside contig boundaries at line %s" % str(c))
            else:
                logging.critical("Error: Genome filter BED contig id not found in reference at line %s" % str(c))
                raise TypeError("Error: Genome filter BED contig id not found in reference at line %s" % str(c))



"""
Functions for creating VCF file.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
import math
from datetime import date
from natsort import natsorted
from nanovar import __version__


def create_vcf(wk_dir, thres, nn_out, ref_path, read_path, read_name, mm_cmd, contig_len_dict, homo_t, het_t, minlen, depth,
               index2te, nv_cmd, alt_seq):
    rdata = nn_out
    # Calculating number of entries
    t = len(rdata)
    # Set range according to number of queries
    k = range(t)
    # Create last line dummy
    rdata.append('dum\tdum\tdum\tdum\tdum\tdum\tdum\tdum\tdum')
    ref_path = os.path.abspath(ref_path)
    read_path = os.path.abspath(read_path)
    vcf_total = open(os.path.join(wk_dir, '%s.nanovar.total.vcf' % read_name), 'w')
    vcf_pass = open(os.path.join(wk_dir, '%s.nanovar.pass.vcf' % read_name), 'w')
    add_header(vcf_total, read_path, ref_path, mm_cmd, read_name, contig_len_dict, thres, depth, nv_cmd)
    add_header(vcf_pass, read_path, ref_path, mm_cmd, read_name, contig_len_dict, thres, depth, nv_cmd)
    sv_sup = open(os.path.join(wk_dir, 'sv_support_reads.tsv'), 'w')
    sv_sup.write('SV-ID\tSupporting_reads (readname~index1,readname~index2...)\n')
    tmpread = []
    out = []
    for i in k:
        tmpread.append(rdata[i])
        covl = int(tmpread[0].split('\t')[10])
        normcov = int(float(tmpread[0].split('\t')[12]))
        dnnscore = float(tmpread[0].split('\t')[13])
        dnn = str(round(float(dnnscore), 3))
        phred = phredc(dnnscore)
        sv_id = tmpread[0].split('\t')[6].split('~')[0]
        bp_name = tmpread[0].split('\t')[3].split(' ')[0]
        chrm1 = tmpread[0].split('\t')[6].split('~')[1].split(':')[0]
        filt = filterer(float(phred), thres)
        phred = str(abs(round(phred, 1)))
        dp = str(covl + normcov)
        ratio = round(float(covl) / int(dp), 3)
        geno = genotyper(ratio, homo_t, het_t)
        if bp_name == 'Nov_Ins':
            sv = '<INS>'
            new_sv_id = 'NV.INS.' + sv_id.split('nv_')[1]
            sv_len = tmpread[0].split('\t')[3].split(' ')[1].split('~')[0]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            if sv_id in index2te:
                te = ";TE=" + ','.join(sorted(index2te[sv_id]))
            else:
                te = ''
            ref_bases = alt_seq[sv_id][0]
            alt_bases = alt_seq[sv_id]
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\t{}\t{}\t'.format(ref_bases, alt_bases) + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=INS;END=' + str(coord1 + 1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(
                covl) + ';NN=' +
                       str(dnn) + te + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'E-Nov_Ins_bp' or bp_name == 'S-Nov_Ins_bp':
            pass
            sv = '<INS>'
            new_sv_id = 'NV.INS.' + sv_id.split('nv_')[1]
            sv_len = tmpread[0].split('\t')[3].split(' ')[1].split('~')[0]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            if sv_id in index2te:
                te = ";TE=" + ','.join(sorted(index2te[sv_id]))
            else:
                te = ''
            ref_bases = alt_seq[sv_id][0]
            alt_bases = alt_seq[sv_id]
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\t{}\t{}\t'.format(ref_bases, alt_bases) + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=INS;END=' + str(coord1 + 1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(
                covl) + ';NN=' +
                       str(dnn) + te + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'Del':
            sv = '<DEL>'
            new_sv_id = 'NV.DEL.' + sv_id.split('nv_')[1]
            sv_len = int(tmpread[0].split('\t')[3].split(' ')[1].split('~')[0])
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[1])
            if coord2 - coord1 < minlen:
                mid = (coord2 + coord1) / 2
                coord1 = max(1, int(mid - round(sv_len/2, 0)))
                coord2 = int(mid + round(sv_len/2, 0) + 1)
            sv_len = '-' + str(coord2 - coord1)
            ref_bases = alt_seq[sv_id]
            alt_bases = alt_seq[sv_id][0]
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\t{}\t{}\t'.format(ref_bases, alt_bases) + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=DEL;END=' + str(coord2) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'Inv':
            sv = '<INV>'
            new_sv_id = 'NV.INV.' + sv_id.split('nv_')[1]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[1])
            if coord2 - coord1 < minlen:
                mid = (coord2 + coord1) / 2
                coord1 = max(1, int(mid - round(minlen / 2, 0)))
                coord2 = int(mid + round(minlen / 2, 0) + 1)
            sv_len = str(coord2 - coord1)
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\tN\t' + str(sv) + '\t' + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=INV;END=' + str(coord2) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'Inv(1)' or bp_name == 'Inv(2)':
            sv = '<INV>'
            new_sv_id = 'NV.INV.' + sv_id.split('nv_')[1]
            # sv_len = tmpread[0].split('\t')[3].split(' ')[1].split('~')[0]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[1])
            if coord2 - coord1 < minlen:
                mid = (coord2 + coord1) / 2
                coord1 = max(1, int(mid - round(minlen / 2, 0)))
                coord2 = int(mid + round(minlen / 2, 0) + 1)
            sv_len = str(coord2 - coord1)
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\tN\t' + str(sv) + '\t' + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=INV;END=' + str(coord2) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'TDupl':
            sv = '<DUP>'
            new_sv_id = 'NV.DUP.' + sv_id.split('nv_')[1]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[1])
            if coord2 - coord1 < minlen:
                mid = (coord2 + coord1) / 2
                coord1 = max(1, int(mid - round(minlen / 2, 0)))
                coord2 = int(mid + round(minlen / 2, 0) + 1)
            sv_len = str(coord2 - coord1)
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\tN\t' + str(sv) + '\t' + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=DUP;END=' + str(coord2) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'Intra-Ins':
            new_sv_id = 'NV.BND.' + sv_id.split('nv_')[1]
            sv_len = '0'
            strands = tmpread[0].split('\t')[3].split(' ')[1].split('~')[1].split(',')
            chrm2 = tmpread[0].split('\t')[6].split('~')[1].split(':')[0]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[1])
            a, b = breakend_alt(strands, chrm1, coord1, chrm2, coord2)
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + ';SV2=TRA' + ';CHR2=' + str(chrm2) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TRA' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm2) + '\t' + str(coord2) + '\t' + str(sv_id) + '\tN\t' + str(b) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord2+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TRA' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'Intra-Ins(1)' or bp_name == 'Intra-Ins(2)':
            new_sv_id = 'NV.BND.' + sv_id.split('nv_')[1]
            sv_len = tmpread[0].split('\t')[3].split(' ')[1].split('~')[0]
            strands = tmpread[0].split('\t')[3].split(' ')[1].split('~')[1].split(',')
            chrm2 = tmpread[0].split('\t')[6].split('~')[1].split(':')[0]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1].split('-')[1])
            a, b = breakend_alt(strands, chrm1, coord1, chrm2, coord2)
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + ';SV2=TPO' + ';CHR2=' + str(chrm2) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TPO' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm2) + '\t' + str(coord2) + '\t' + str(sv_id) + '\tN\t' + str(b) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord2+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TPO' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'Inter-Ins(1)' or bp_name == 'Inter-Ins(2)':
            new_sv_id = 'NV.BND.' + sv_id.split('nv_')[1]
            sv_len = tmpread[0].split('\t')[3].split(' ')[1].split('~')[0]
            strands = tmpread[0].split('\t')[3].split(' ')[1].split('~')[1].split(',')
            chrm2 = tmpread[0].split('\t')[6].split('~')[2].split(':')[0]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[2].split(':')[1])
            a, b = breakend_alt(strands, chrm1, coord1, chrm2, coord2)
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + ';SV2=TPO' + ';CHR2=' + str(chrm2) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TPO' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm2) + '\t' + str(coord2) + '\t' + str(sv_id) + '\tN\t' + str(b) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord2+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TPO' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        elif bp_name == 'InterTx':
            new_sv_id = 'NV.BND.' + sv_id.split('nv_')[1]
            sv_len = '0'
            strands = tmpread[0].split('\t')[3].split(' ')[1].split('~')[1].split(',')
            chrm2 = tmpread[0].split('\t')[6].split('~')[2].split(':')[0]
            coord1 = int(tmpread[0].split('\t')[6].split('~')[1].split(':')[1])
            coord2 = int(tmpread[0].split('\t')[6].split('~')[2].split(':')[1])
            a, b = breakend_alt(strands, chrm1, coord1, chrm2, coord2)
            out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(new_sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
                       filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
                       str(dnn) + ';SV2=TRA' + ';CHR2=' + str(chrm2) + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm1) + '\t' + str(coord1) + '\t' + str(sv_id) + '\tN\t' + str(a) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord1+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TRA' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
            # out.append(str(chrm2) + '\t' + str(coord2) + '\t' + str(sv_id) + '\tN\t' + str(b) + '\t' + str(phred) + '\t' +
            #            filt + '\t' + 'SVTYPE=BND;END=' + str(coord2+1) + ';SVLEN=' + str(sv_len) + ';SR=' + str(covl) + ';NN=' +
            #            str(dnn) + ';SV2=TRA' + '\tGT:DP:AD\t' + geno + ':' + dp + ':' + str(normcov) + ',' + str(covl))
        else:
            raise Exception("Error: Unrecognised breakpoint name")
        sv_sup.write(new_sv_id + '\t' + tmpread[0].split('\t')[8] + ',' + tmpread[0].split('\t')[11] + '\n')
        tmpread = []
    out_sort1 = natsorted(out, key=lambda x: x.split('\t')[1])  # sort by POS
    total_vcf = natsorted(out_sort1, key=lambda x: x.split('\t')[0])  # sort by CHROM
    for e in total_vcf:
        vcf_total.write(e + '\n')
        if e.split('\t')[6] == 'PASS':
            vcf_pass.write(e + '\n')
    vcf_total.close()
    vcf_pass.close()
    sv_sup.close()


# Write VCF header
def add_header(vcf_file, read_path, ref_path, mm_cmd, read_name, contig_len_dict, thres, depth, nv_cmd):
    today = date.today()
    today_date = today.strftime("%d-%m-%Y")
    vcf_file.write('##fileformat=VCFv4.2\n')
    vcf_file.write('##fileDate=%s\n' % today_date)
    vcf_file.write('##source=NanoVar-%s\n' % __version__)
    vcf_file.write('##source_reads=%s\n' % read_path)
    vcf_file.write('##reference=%s\n' % ref_path)
    vcf_file.write('##command=%s\n' % nv_cmd)
    vcf_file.write('##mapping=%s\n' % mm_cmd)
    vcf_file.write('##phasing=none\n')
    vcf_file.write('##depth_of_coverage=%sX\n' % str(depth))
    for key in contig_len_dict:
        vcf_file.write('##contig=<ID=%s,length=%s>\n' % (key, str(contig_len_dict[key])))
    vcf_file.write('##ALT=<ID=INS,Description="Insertion of novel sequence relative to the reference">\n')
    vcf_file.write('##ALT=<ID=DEL,Description="Deletion relative to the reference">\n')
    vcf_file.write('##ALT=<ID=INV,Description="Inversion relative to the reference">\n')
    vcf_file.write('##ALT=<ID=DUP,Description="Tandem duplication relative to the reference">\n')
    vcf_file.write('##ALT=<ID=BND,Description="Breakend relative to the reference">\n')
    vcf_file.write('##FILTER=<ID=PASS,Description="Quality above/equal to %s">\n' % str(thres))
    vcf_file.write('##FILTER=<ID=FAIL,Description="Quality below %s">\n' % str(thres))
    vcf_file.write('##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant">\n')
    vcf_file.write('##INFO=<ID=END,Number=1,Type=Integer,Description="End position of the described variant">\n')
    vcf_file.write('##INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Estimated length of the variant">\n')
    vcf_file.write('##INFO=<ID=SR,Number=1,Type=Integer,Description="Number of supporting reads">\n')
    vcf_file.write('##INFO=<ID=NN,Number=1,Type=Float,Description="Neural network confidence probability">\n')
    vcf_file.write('##INFO=<ID=SV2,Number=1,Type=String,Description="BND SV assessment: TPO - Transposition, '
                   'TRA - Translocation">\n')
    vcf_file.write('##INFO=<ID=TE,Number=.,Type=String,Description="Identity of transposable element insertion">\n')
    vcf_file.write('##INFO=<ID=CHR2,Number=1,Type=String,Description="Second chromosome name for BND SVs">\n')
    vcf_file.write('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\n')
    vcf_file.write('##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read depth">\n')
    vcf_file.write('##FORMAT=<ID=AD,Number=R,Type=Integer,Description="Read depth for each allele">\n')
    vcf_file.write('#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n' % read_name)


# Phred quality calculator
def phredc(p):
    probfalse = 1.000000001 - float(p)
    return math.log10(probfalse)*(-10)


# Filter based on threshold score
def filterer(x, thres):
    if x < thres:
        return "FAIL"
    else:
        return "PASS"


# Predict SV genotype
def genotyper(x, homo_t, het_t):
    if x >= homo_t:
        return "1/1"
    elif x >= het_t:
        return "0/1"
    else:
        return "./."


# Construct breakend alteration
def breakend_alt(strand_list, chrm1, coord1, chrm2, coord2):
    coord1 = str(coord1)
    coord2 = str(coord2)
    if strand_list == ['+', '+']:
        a = 'N[' + chrm2 + ':' + coord2 + '['
        b = ']' + chrm1 + ':' + coord1 + ']N'
    elif strand_list == ['+', '-']:
        a = 'N]' + chrm2 + ':' + coord2 + ']'
        b = 'N]' + chrm1 + ':' + coord1 + ']'
    elif strand_list == ['-', '+']:
        a = '[' + chrm2 + ':' + coord2 + '[N'
        b = '[' + chrm1 + ':' + coord1 + '[N'
    elif strand_list == ['-', '-']:
        a = ']' + chrm2 + ':' + coord2 + ']N'
        b = 'N[' + chrm1 + ':' + coord1 + '['
    else:
        raise Exception('Strand information not found.')
    return a, b
